!!$=head1 NAME
!!$
!!$Born modeling Adjoint (RTM) - Acoustic RTM operator with constant density. (Version in which source wavefield can be read from file)
!!$
!!$=head1 SYNOPSIS
!!$
!!$Bornfast_adj.x <velocity.H deltadata=data.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefields=wavefields.H wavefieldr=wavefieldr.H >deltamodel.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation Adjoint Born modeling for acoustic approximation with absorbing boundary (i.e., RTM). It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model, the perturbation in the data and return the perturbed model. !NOTE= it assumes that we can keep in memory the source wavefield at source-wavelet sampling!
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<velocity.H> - Velocity model in SEP3D format (theoretically [m/s]). (Constructed using Velpar and/or Vel)
!!$
!!$=item from par
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<deltadata> - Perturbed data.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefields.
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save (wantwavefield=1) or read (readwavefield=1) source wavefield
!!$
!!$B<filename> I<wavefieldr> - [wavefieldr.H] File where to save receiver wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<readwavefield> - [0] If 1 the program reads the source wavefield from file.
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<vel_padded> - [0] Velocity model was padded (100 elements), useful for FWI
!!$
!!$B<integer> I<sqslow> - [0] If 1 output model perturbation is slowness square; otherwise, velocity perturbation is assumed.
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Born_adj
  
  !Used modules 
  use sep
  use read_pos
  use sec_deriv
  use Pad_vel
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_step
  use weight_sec
  use Pad_pert
  use sec_deriv_field


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: vel   !velocity model 
  real, dimension(:,:,:), allocatable					:: model   !velocity perturbation 
  real, dimension(:), allocatable						:: wavelet, wavelet_buff !Source wavelet 
  integer												:: wantwavefield, readwavefield, verbose, sqslow
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right
  real, dimension(:,:,:), allocatable					:: sec_scale 
  real, dimension(:,:,:,:), allocatable					:: source_wavefield 
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold, ptemp
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: vel_pad !Padded velocity model
  real, dimension(:,:,:), allocatable					:: scat_sou_right !for imaging condition
  real													:: dt !Temporal sampling of the propagator
  real, dimension(:), allocatable						:: wavelet_fine, scaled_wavelet !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it, it2, itw !Cycles variables
  integer												:: iz, ix, iy !#######Cycles variables for fixing problem with ifort########
  integer												:: wc, srite, rc, sreed, sk, sseek_block
  integer												:: desample, vel_padded
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr, total_time
  integer												:: count1, count2,ierr, icount, num_threads
  real*8, dimension(:,:), allocatable					:: interp_filter
  integer*8												:: rbit !necessary for reading big files
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
		  
  call from_par ("verbose",verbose,0)

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)"####################Acoustic 2D/3D wave propagator#####################"
  if(verbose)write(0,*)"####################  Adjoint Born modeling (RTM)  ####################"  
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("readwavefield",readwavefield,0)
  call from_par ("vel_padded",vel_padded,0)
  call from_par ("desample",desample,1)
  call from_par ("dt",dt,0.)
  call from_par ("sqslow",sqslow,0)
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the velocity and the perturbation
  if(vel_padded .eq. 0) then
  	allocate(vel(nz,nx,ny))
  	vel(:,:,:) = 0.
  	rc=sreed("in" , vel , nz*nx*ny*4)
  	!Adding absorbing boundaries to the velocity model
  	call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	allocate(vel_pad(nz,nx,ny))
  	vel_pad(:,:,:) = 0.
  	rc=sreed("in" , vel_pad, nz*nx*ny*4)
  endif
  
  allocate(model(nz,nx,ny))
  model(:,:,:) = 0.
  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minval(vel_pad)/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxval(vel_pad)/min(dx,dy,dz)
  if(verbose)write(0,*) ""


  !Computing the second order derivative on the source wavelet
  !allocate(wavelet_buff(nw))
  !call sec_deriv_init(nw, dw)
  call sec_deriv_field_init(nw, dw)
  !call sec_deriv_op(.false., .false., wavelet, wavelet_buff)
  !wavelet = wavelet_buff

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_step_init(nz, nx, ny, nt, nv_pad, vel_pad, dt, dz, dx, dy,twoD)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,vel_pad,twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  allocate(scat_sou_right(nz,nx,ny),sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = -dt*dt*vel_pad*vel_pad
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !allocating memory for source wavefield
  allocate(source_wavefield(nz,nx,ny,nw))
  source_wavefield = 0.
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))

  
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  
  if (wantwavefield) then
	if(verbose)write(0,*)"#Program will be writing wavefield as requested#"
	do is = 1,ns
		do it = 1, nw
			wc=srite("wavefieldr",pcur,nz*nx*ny*4)
		enddo
	enddo
  endif
  
  total_time = 0.
  rbit = nz*nx*ny
  rbit = rbit*nw*4
  do is = 1,ns
 	!moving the file writing position with sseek
  	if (wantwavefield) sk = sseek_block("wavefieldr",nw*(is),4*nz*nx*ny,0)
  	old = 0.; cur = 0.
  	source_wavefield = 0.
  	sec_source_left=0.; sec_source_right=0.
  	scat_sou_right = 0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	call system_clock(count_rate=cr)
	call system_clock(count1)
	if(.not. readwavefield) then
		!Scaling the source
		scaled_wavelet = wavelet_fine
		call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
		call Sou_scale_op(scaled_wavelet,vel_pad) !Adjoint = Forward (input-output same variable)
		if(verbose)write(0,*)"Propagating source wavefield..."
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
		endif
		do itw = 1, nw-1 !Forward
			it = (itw-1)*desample
			!Propagate source field	
			do it2 = 1, desample
				call Prop_step_op(.false., pold, pcur, pnew)
				!Adding source (perturbed and source wavefield)
				if (twoD) then
					call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				else
					call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				endif	
				!Apply adjoint(fine to coarse) of linear interpolation
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,itw) = source_wavefield(iz,ix,iy,itw) + pcur(iz,ix,iy)*interp_filter(it2,1)
							source_wavefield(iz,ix,iy,itw+1) = source_wavefield(iz,ix,iy,itw+1) + pcur(iz,ix,iy)*interp_filter(it2,2)	
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
				!Swapping pointers
				pold => pcur 
				pcur => pnew
				pnew => pold
			enddo
			!saving source wavefield
			if (wantwavefield) then
				wc=srite("wavefields",source_wavefield(:,:,:,itw),nz*nx*ny*4)
			endif
		
			if (mod(itw,(nw/10))==0) then
        		if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        		close(0)
      		endif
      	
		enddo
		!Last step of the source wavefield just saving source wavefield
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,nw) = source_wavefield(iz,ix,iy,nw) + pcur(iz,ix,iy)*interp_filter(1,1)
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,nw),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
		if(verbose)write(0,'(a)',advance="yes")""
		if(verbose)write(0,*)"End Propagation source wavefield"
	else	
		if(verbose)write(0,*)"Reading source wavefield"
		rc=sreed("wavefields",source_wavefield,rbit)
	endif
	
	!Receiver wavefield and imaging condition
	if(verbose)write(0,*)"Propagating receiver wavefield..."
	icount = 0.
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
	!Creating "adjoint" source at the receivers
	if (dw .le. dt) then
		do ig = 1,ng
			rc=sreed("deltadata",recfield,nw*4)
			recfield_fine(:,ig) = recfield 
		enddo
	else
		do ig = 1,ng
			rc=sreed("deltadata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_fine(:,ig))
		enddo
	endif
	!zero-out the model
	model(:,:,:) = 0.
	do itw = nw, 2, -1 !Adjoint 
		!Propagate scattered wavefield
		it = (itw-2)*desample
		do it2 = desample+1, 2, -1
			call Prop_step_op(.true., poldpert, pcurpert, pnewpert)
			!Adding receiver deltadata 
			if (twoD) then
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
				enddo
			else
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
				enddo
			endif
			!Apply adjoint(fine to coarse) of linear interpolation 
			!####ADDED LOOPS BECAUSE PROBLEM OF IFORT ON SEP MACHINE######
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						psec_source_left(iz,ix,iy) = psec_source_left(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(it2,1)
						psec_source_right(iz,ix,iy) = psec_source_right(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(it2,2)
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
			
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		
		!Scale secondary source (i.e., back-propagated data)
		scat_sou_right(:,:,:)=psec_source_right(:,:,:)*sec_scale(:,:,:)
		call weight_sec_op(.true., .false., psec_source_right(:,:,:), scat_sou_right(:,:,:))
		!Apply imaging condition 
		call sec_deriv_field_op(source_wavefield, scat_sou_right, itw)
		model(:,:,:) = model(:,:,:) + psec_source_right(:,:,:) * scat_sou_right(:,:,:)
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek_block("wavefieldr",-4,nz*nx*ny,1)
			wc=srite("wavefieldr",psec_source_right(:,:,:),nz*nx*ny*4)
			sk = sseek_block("wavefieldr",-4,nz*nx*ny,1)
		endif
		icount = icount + 1
		if (mod(icount,(nw/10))==0) then
	        	if(verbose)write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
	        	close(0)
		endif	
		
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_left = 0.
	enddo
	!Last step (already propagated)
	itw = 1
	!####ADDED LOOPS BECAUSE PROBLEM OF IFORT ON SEP MACHINE######
	!$OMP PARALLEL DO private(iz,ix,iy)
	do iy=1,ny
		do ix=1,nx
			do iz=1,nz
				psec_source_right(iz,ix,iy) = psec_source_right(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(1,2)
			enddo
		enddo
	enddo
	!$OMP END PARALLEL DO
	!Scale secondary source (i.e., back-propagated data)
		scat_sou_right(:,:,:)=psec_source_right(:,:,:)*sec_scale(:,:,:)
		call weight_sec_op(.true., .false., psec_source_right(:,:,:), scat_sou_right(:,:,:))
		!Apply imaging condition 
		call sec_deriv_field_op(source_wavefield, scat_sou_right, itw)
		model(:,:,:) = model(:,:,:) + psec_source_right(:,:,:) * scat_sou_right(:,:,:)
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek_block("wavefieldr",-4,nz*nx*ny,1)
			wc=srite("wavefieldr",psec_source_right(:,:,:),nz*nx*ny*4)
			sk = sseek_block("wavefieldr",-4,nz*nx*ny,1)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	total_time = total_time + (count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing model for source n°: ", is
	!Write model for is source
	!compute velocity perturbation from model perturbation
  	if(sqslow .eq. 0) model(:,:,:) = -2.0*model(:,:,:)/(vel_pad(:,:,:)*vel_pad(:,:,:)*vel_pad(:,:,:))
	wc=srite("out",model(:,:,:),nz*nx*ny*4)
   	if(verbose)write(0,*) ""

  enddo
  
  if(verbose)write(0,*) "Total propagation time: ", total_time ," seconds"
  if(verbose)write(0,*) "Total propagation time: ", total_time/60.0 ," minutes"
  
  !Closing used routines
  call Prop_step_close()
  call weight_sec_close()

  !Writing information to headers
  call to_history("n1",nz)
  call to_history("o1",oz)
  call to_history("d1",dz)
  call to_history("label1","z [m]")
  call to_history("n2",nx)
  call to_history("o2",ox)
  call to_history("d2",dx)
  call to_history("label2","x [m]")
  call to_history("n3",ny)
  call to_history("o3",oy)
  call to_history("d3",dy)
  call to_history("label3","y [m]")
  call to_history("n4",ns)
  call to_history("o4",1.)
  call to_history("d4",1.)
  call to_history("label4","Source number")
  
  
  if (wantwavefield) then
  
  	if(twoD) then
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
  			call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",nt/desample+1,"wavefields")
  			call to_history("o3",ow,"wavefields")
 		 	call to_history("d3",dt*desample,"wavefields")
  			call to_history("label3","Time [s]","wavefields")
  			call to_history("n4",ns,"wavefields")
  			call to_history("o4",1.,"wavefields")
  			call to_history("d4",1.,"wavefields")
  			call to_history("label4","Source number","wavefields") 
  		endif 
  		
  		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
		call to_history("n3",nt/desample+1,"wavefieldr")
  		call to_history("o3",ow,"wavefieldr")
 	 	call to_history("d3",dt*desample,"wavefieldr")
  		call to_history("label3","Time [s]","wavefieldr")
  		call to_history("n4",ns,"wavefieldr")
  		call to_history("o4",1.,"wavefieldr")
  		call to_history("d4",1.,"wavefieldr")
  		call to_history("label4","Source number","wavefieldr")  		
  	else
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
	  		call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",ny,"wavefields")
		  	call to_history("o3",oy,"wavefields")
		  	call to_history("d3",dy,"wavefields")
  			call to_history("label3","y [m]","wavefields")
			call to_history("n4",nt/desample+1,"wavefields")
  			call to_history("o4",ow,"wavefields")
 		 	call to_history("d4",dt*desample,"wavefields")
  			call to_history("label4","Time [s]","wavefields")
  			call to_history("n5",ns,"wavefields")
  			call to_history("o5",1.,"wavefields")
  			call to_history("d5",1.,"wavefields")
  			call to_history("label5","Source number","wavefields")
  		endif
  		
  		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
		call to_history("n3",ny,"wavefieldr")
	  	call to_history("o3",oy,"wavefieldr")
	  	call to_history("d3",dy,"wavefieldr")
  		call to_history("label3","y [m]","wavefieldr")
		call to_history("n4",nt/desample+1,"wavefieldr")
  		call to_history("o4",ow,"wavefieldr")
 	 	call to_history("d4",dt*desample,"wavefieldr")
  		call to_history("label4","Time [s]","wavefieldr")
  		call to_history("n5",ns,"wavefieldr")
  		call to_history("o5",1.,"wavefieldr")
  		call to_history("d5",1.,"wavefieldr")
  		call to_history("label5","Source number","wavefieldr")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"

  !Deallocating memory
  deallocate(interp_filter,source_wavefield)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale,scat_sou_right)
  deallocate(cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(model,vel_pad)
  deallocate(wavelet)
  !deallocate(wavelet_buff,wavelet)
  call sep_close()
  call exit (0)
end program Born_adj

