!Module for computing the C5_acoustic operator for full acoustic equation 
!Model dimension (z,x,y) Source wavefield
!Data dimension (z,x,y) Wavefield
!D(z,x,y) = C5_acoustic(Model(z,x,y))
!Model(z,x,y) = C5_acoustic*(D(z,x,y))
module C5_acoustic

  use C2_acoustic
  use weight
  use G

  implicit none
  !Module variables
  integer, private									:: nz, nx, ny, nt
  real, private										:: dt
  real, dimension(:,:,:), allocatable, private		:: tmp, tmp1, scale

  contains


	!Initialization routine
	subroutine C5_acoustic_init(nz_in, nx_in, ny_in, bulk_in, rho_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, dimension(:,:,:), pointer			:: bulk_in, rho_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dt = dt_in

		if (.not. allocated(tmp)) allocate(tmp(nz,nx,ny))
		tmp = 0.
		if (.not. allocated(tmp1)) allocate(tmp1(nz,nx,ny))
		tmp1 = 0.
		if (.not. allocated(scale)) allocate(scale(nz,nx,ny))
		scale = sqrt(bulk_in(:,:,:)/rho_in(:,:,:))*dt
		
		!Initializing other modules
		call C2_acoustic_init(nz, nx, ny, bulk_in, rho_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		

		return
	end subroutine C5_acoustic_init


  	subroutine C5_acoustic_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:)  :: mm,dd
    		call C5_acoustic_op(adj,add,mm,dd)
  	end subroutine C5_acoustic_call

	
	!Operator routine
	subroutine C5_acoustic_op(adj, add, mm, dd)
	
		!Input-Output variables	
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables

    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model

			!Operator: (I-G*v*dt)w
			call weight_op(.true.,.false.,tmp,dd(:,:,:))
			call G_op(.true.,.false.,tmp1,(tmp(:,:,:)*scale(:,:,:)))
			mm(:,:,:) = mm(:,:,:) + tmp(:,:,:) - tmp1(:,:,:)
			!Operator: (lapla*v^2*dt^2+2I)(I-w)
			call weight_op(.true.,.false.,tmp,dd(:,:,:))
			tmp(:,:,:) = dd(:,:,:) - tmp(:,:,:)
			call C2_acoustic_op(.true.,.false.,tmp1,tmp)
			mm(:,:,:) = mm(:,:,:) + tmp1(:,:,:)

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data

			!Operator: (I-w)(v^2*dt^2*lapla+2I)
			call C2_acoustic_op(.false.,.false.,mm(:,:,:),tmp)
			call weight_op(.false.,.false.,tmp,tmp1)
			dd(:,:,:) = dd(:,:,:) + tmp(:,:,:) - tmp1(:,:,:)
			!!Operator: w(I-v*dt*G)
			call G_op(.false.,.false.,mm(:,:,:),tmp1(:,:,:))
			call weight_op(.false.,.false.,(mm(:,:,:)-scale(:,:,:)*tmp1(:,:,:)),tmp(:,:,:))
			dd(:,:,:) = dd(:,:,:) + tmp(:,:,:)
		endif

		return
	end subroutine C5_acoustic_op


	!Subroutine for deallocating variables
	subroutine C5_acoustic_close()
		if (allocated(tmp)) deallocate(tmp)
		if (allocated(tmp1)) deallocate(tmp1)
		if (allocated(scale)) deallocate(scale)
		call weight_close()
	end subroutine C5_acoustic_close


end module C5_acoustic
