!Module for computing the div(1/rho grad) with Mattsson's compact stencil (zero-outside -> Forward=Adjoint)
!6th-order accurate of div(1/rho grad)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput Db2 of wavefield
!D(z,x,y) = Db2(Model(z,x,y))
!Model(z,x,y) = Db2*(D(z,x,y))
!For reference see:
!Ken Mattsson, 2012, Summation by Parts Operators for Finite Difference Approximations of Second-Derivatives with Variable Coefficients, J. Sci. Comput., 51, pp. 650-682

module Db2

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  real, private									:: dz, dx, dy
  real, private, dimension(:,:,:),allocatable 	:: C0, Cz1r, Cz2r, Cz3r, Cz1l, Cz2l, Cz3l
  real, private, dimension(:,:,:),allocatable 	:: Cx1r, Cx2r, Cx3r, Cx1l, Cx2l, Cx3l
  real, private, dimension(:,:,:),allocatable 	:: Cy1r, Cy2r, Cy3r, Cy1l, Cy2l, Cy3l
  real, private, dimension(:,:,:),allocatable	:: tmp
  real, private, dimension(:,:,:),allocatable	:: b ! 1/rho
  logical, private								:: twoD

  contains


	!Initialization routine
	subroutine Db2_init(nz_in, nx_in, ny_in, dz_in, dx_in, dy_in, rho_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, intent(in)						:: dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		real, dimension(:,:,:), intent(in)		:: rho_in
		!Internal variables
		integer 								:: buffer, iz, ix, iy
		real									:: idz, idx, idy

   		parameter(buffer=3)

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dz = dz_in; idz = 1./(dz*dz)
		dx = dx_in; idx = 1./(dx*dx)
		dy = dy_in; idy = 1./(dy*dy)
		twoD = twoD_in

        if (.not. allocated(tmp)) then
        	!temporary array for 1/rho
        	allocate(b(nz,nx,ny))
        	b(:,:,:) = 1./rho_in(:,:,:) 
 			
			if (.not. twoD) then
				allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer))	
 				allocate(C0(nz,nx,ny),Cz1r(nz,nx,ny),Cz2r(nz,nx,ny),Cz3r(nz,nx,ny),Cz1l(nz,nx,ny),Cz2l(nz,nx,ny),Cz3l(nz,nx,ny))
 				allocate(Cx1r(nz,nx,ny),Cx2r(nz,nx,ny),Cx3r(nz,nx,ny),Cx1l(nz,nx,ny),Cx2l(nz,nx,ny),Cx3l(nz,nx,ny))
				allocate(Cy1r(nz,nx,ny),Cy2r(nz,nx,ny),Cy3r(nz,nx,ny),Cy1l(nz,nx,ny),Cy2l(nz,nx,ny),Cy3l(nz,nx,ny))
				C0 = 0.
 				Cz1r = 0.; Cz1l = 0.; Cz2r = 0.; Cz2l = 0.; Cz3r = 0.; Cz3l = 0.;
 				Cx1r = 0.; Cx1l = 0.; Cx2r = 0.; Cx2l = 0.; Cx3r = 0.; Cx3l = 0.;
 				Cy1r = 0.; Cy1l = 0.; Cy2r = 0.; Cy2l = 0.; Cy3r = 0.; Cy3l = 0.;
				!Initializing 6th-order FD div(1/rho grad) 3D
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = buffer+1, ny-buffer
					do ix = buffer+1, nx-buffer
						do iz = buffer+1, nz-buffer
							Cy3l(iz,ix,iy) = -(1./40.*(b(iz,ix,iy-2)+b(iz,ix,iy-1)) -11./360.*(b(iz,ix,iy-3)+b(iz,ix,iy)))*idy
							Cx3l(iz,ix,iy) = -(1./40.*(b(iz,ix-2,iy)+b(iz,ix-1,iy)) -11./360.*(b(iz,ix-3,iy)+b(iz,ix,iy)))*idx
							Cz3l(iz,ix,iy) = -(1./40.*(b(iz-2,ix,iy)+b(iz-1,ix,iy)) -11./360.*(b(iz-3,ix,iy)+b(iz,ix,iy)))*idz
							Cy2l(iz,ix,iy) = -(1./20.*(b(iz,ix,iy-3)+b(iz,ix,iy+1)) -3./10.*(b(iz,ix,iy-1)) +7./40.*(b(iz,ix,iy)+b(iz,ix,iy-2)))*idy
							Cx2l(iz,ix,iy) = -(1./20.*(b(iz,ix-3,iy)+b(iz,ix+1,iy)) -3./10.*(b(iz,ix-1,iy)) +7./40.*(b(iz,ix,iy)+b(iz,ix-2,iy)))*idx
							Cz2l(iz,ix,iy) = -(1./20.*(b(iz-3,ix,iy)+b(iz+1,ix,iy)) -3./10.*(b(iz-1,ix,iy)) +7./40.*(b(iz,ix,iy)+b(iz-2,ix,iy)))*idz
							Cy1l(iz,ix,iy) = -(-1./40.*(b(iz,ix,iy-3)+b(iz,ix,iy+2)) -3./10.*(b(iz,ix,iy-2)+b(iz,ix,iy+1)) -17./40*(b(iz,ix,iy)+b(iz,ix,iy-1)))*idy
							Cx1l(iz,ix,iy) = -(-1./40.*(b(iz,ix-3,iy)+b(iz,ix+2,iy)) -3./10.*(b(iz,ix-2,iy)+b(iz,ix+1,iy)) -17./40*(b(iz,ix,iy)+b(iz,ix-1,iy)))*idx
							Cz1l(iz,ix,iy) = -(-1./40.*(b(iz-3,ix,iy)+b(iz+2,ix,iy)) -3./10.*(b(iz-2,ix,iy)+b(iz+1,ix,iy)) -17./40*(b(iz,ix,iy)+b(iz-1,ix,iy)))*idz
							C0(iz,ix,iy) = -((idz+idx+idy)*101.0/180.0*b(iz,ix,iy) &
									  +1./180.0*((b(iz-3,ix,iy)+b(iz+3,ix,iy))*idz+(b(iz,ix-3,iy)+b(iz,ix+3,iy))*idx+(b(iz,ix,iy-3)+b(iz,ix,iy+3))*idy) &
									  +1./8.0*((b(iz-2,ix,iy)+b(iz+2,ix,iy))*idz+(b(iz,ix-2,iy)+b(iz,ix+2,iy))*idx+(b(iz,ix,iy-2)+b(iz,ix,iy+2))*idy) &
									  +19./20.0*((b(iz-1,ix,iy)+b(iz+1,ix,iy))*idz+(b(iz,ix-1,iy)+b(iz,ix+1,iy))*idx+(b(iz,ix,iy-1)+b(iz,ix,iy+1))*idy))
							Cz1r(iz,ix,iy) = -(-1./40.*(b(iz-2,ix,iy)+b(iz+3,ix,iy)) -3./10.*(b(iz-1,ix,iy)+b(iz+2,ix,iy)) -17./40*(b(iz,ix,iy)+b(iz+1,ix,iy)))*idz
							Cx1r(iz,ix,iy) = -(-1./40.*(b(iz,ix-2,iy)+b(iz,ix+3,iy)) -3./10.*(b(iz,ix-1,iy)+b(iz,ix+2,iy)) -17./40*(b(iz,ix,iy)+b(iz,ix+1,iy)))*idx
							Cy1r(iz,ix,iy) = -(-1./40.*(b(iz,ix,iy-2)+b(iz,ix,iy+3)) -3./10.*(b(iz,ix,iy-1)+b(iz,ix,iy+2)) -17./40*(b(iz,ix,iy)+b(iz,ix,iy+1)))*idy
							Cz2r(iz,ix,iy) = -(1./20.*(b(iz-1,ix,iy)+b(iz+3,ix,iy)) -3./10.*(b(iz+1,ix,iy)) +7./40.*(b(iz,ix,iy)+b(iz+2,ix,iy)))*idz
							Cx2r(iz,ix,iy) = -(1./20.*(b(iz,ix-1,iy)+b(iz,ix+3,iy)) -3./10.*(b(iz,ix+1,iy)) +7./40.*(b(iz,ix,iy)+b(iz,ix+2,iy)))*idx
							Cy2r(iz,ix,iy) = -(1./20.*(b(iz,ix,iy-1)+b(iz,ix,iy+3)) -3./10.*(b(iz,ix,iy+1)) +7./40.*(b(iz,ix,iy)+b(iz,ix,iy+2)))*idy
							Cz3r(iz,ix,iy) = -(1./40.*(b(iz+1,ix,iy)+b(iz+2,ix,iy)) -11./360.*(b(iz,ix,iy)+b(iz+3,ix,iy)))*idz
							Cx3r(iz,ix,iy) = -(1./40.*(b(iz,ix+1,iy)+b(iz,ix+2,iy)) -11./360.*(b(iz,ix,iy)+b(iz,ix+3,iy)))*idx
							Cy3r(iz,ix,iy) = -(1./40.*(b(iz,ix,iy+1)+b(iz,ix,iy+2)) -11./360.*(b(iz,ix,iy)+b(iz,ix,iy+3)))*idy
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
				
				!Setting the filters in the buffer region
				!$OMP PARALLEL DO private(ix)
				do ix = 1,buffer
						!Extending in the x
						!LEFT x
				 		Cx3l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx3l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz3l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz3l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cx2l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx2l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz2l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz2l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cx1l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx1l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz1l(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz1l(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		C0(buffer+1:nz-buffer,ix,buffer+1:ny-buffer)   = C0(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz1r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz1r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cx1r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx1r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz2r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz2r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cx2r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx2r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cz3r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cz3r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		Cx3r(buffer+1:nz-buffer,ix,buffer+1:ny-buffer) = Cx3r(buffer+1:nz-buffer,buffer+1,buffer+1:ny-buffer)
				 		!RIGHT x
				 		Cx3l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx3l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz3l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz3l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cx2l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx2l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz2l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz2l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cx1l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx1l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz1l(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz1l(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		C0(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer)   = C0(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz1r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz1r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cx1r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx1r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz2r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz2r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cx2r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx2r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cz3r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cz3r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		Cx3r(buffer+1:nz-buffer,ix+nx-buffer,buffer+1:ny-buffer) = Cx3r(buffer+1:nz-buffer,nx-buffer,buffer+1:ny-buffer)
				 		!Extending in the y
				 		!LEFT y
				 		iy = ix
				 		Cx3l(buffer+1:nz-buffer,:,iy) = Cx3l(buffer+1:nz-buffer,:,buffer+1)
				 		Cz3l(buffer+1:nz-buffer,:,iy) = Cz3l(buffer+1:nz-buffer,:,buffer+1)
				 		Cx2l(buffer+1:nz-buffer,:,iy) = Cx2l(buffer+1:nz-buffer,:,buffer+1)
				 		Cz2l(buffer+1:nz-buffer,:,iy) = Cz2l(buffer+1:nz-buffer,:,buffer+1)
				 		Cx1l(buffer+1:nz-buffer,:,iy) = Cx1l(buffer+1:nz-buffer,:,buffer+1)
				 		Cz1l(buffer+1:nz-buffer,:,iy) = Cz1l(buffer+1:nz-buffer,:,buffer+1)
				 		C0(buffer+1:nz-buffer,:,iy)   = C0(buffer+1:nz-buffer,:,buffer+1)
				 		Cz1r(buffer+1:nz-buffer,:,iy) = Cz1r(buffer+1:nz-buffer,:,buffer+1)
				 		Cx1r(buffer+1:nz-buffer,:,iy) = Cx1r(buffer+1:nz-buffer,:,buffer+1)
				 		Cz2r(buffer+1:nz-buffer,:,iy) = Cz2r(buffer+1:nz-buffer,:,buffer+1)
				 		Cx2r(buffer+1:nz-buffer,:,iy) = Cx2r(buffer+1:nz-buffer,:,buffer+1)
				 		Cz3r(buffer+1:nz-buffer,:,iy) = Cz3r(buffer+1:nz-buffer,:,buffer+1)
				 		Cx3r(buffer+1:nz-buffer,:,iy) = Cx3r(buffer+1:nz-buffer,:,buffer+1)
				 		!RIGHT y
				 		Cx3l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx3l(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz3l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz3l(buffer+1:nz-buffer,:,ny-buffer)
				 		Cx2l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx2l(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz2l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz2l(buffer+1:nz-buffer,:,ny-buffer)
				 		Cx1l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx1l(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz1l(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz1l(buffer+1:nz-buffer,:,ny-buffer)
				 		C0(buffer+1:nz-buffer,:,iy+ny-buffer)   = C0(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz1r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz1r(buffer+1:nz-buffer,:,ny-buffer)
				 		Cx1r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx1r(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz2r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz2r(buffer+1:nz-buffer,:,ny-buffer)
				 		Cx2r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx2r(buffer+1:nz-buffer,:,ny-buffer)
				 		Cz3r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cz3r(buffer+1:nz-buffer,:,ny-buffer)
				 		Cx3r(buffer+1:nz-buffer,:,iy+ny-buffer) = Cx3r(buffer+1:nz-buffer,:,ny-buffer)
				 		!Extending in the z
						!UP z
						iz = ix
						Cx3l(iz,:,:) = Cx3l(buffer+1,:,:)
						Cz3l(iz,:,:) = Cz3l(buffer+1,:,:)
						Cx2l(iz,:,:) = Cx2l(buffer+1,:,:)
						Cz2l(iz,:,:) = Cz2l(buffer+1,:,:)
						Cx1l(iz,:,:) = Cx1l(buffer+1,:,:)
						Cz1l(iz,:,:) = Cz1l(buffer+1,:,:)
						C0(iz,:,:) = C0(buffer+1,:,:)
						Cz1r(iz,:,:) = Cz1r(buffer+1,:,:)
						Cx1r(iz,:,:) = Cx1r(buffer+1,:,:)
						Cz2r(iz,:,:) = Cz2r(buffer+1,:,:)
						Cx2r(iz,:,:) = Cx2r(buffer+1,:,:)
						Cz3r(iz,:,:) = Cz3r(buffer+1,:,:)
						Cx3r(iz,:,:) = Cx3r(buffer+1,:,:)
						!DOWN z
						Cx3l(iz+nz-buffer,:,:) = Cx3l(nz-buffer,:,:)
						Cz3l(iz+nz-buffer,:,:) = Cz3l(nz-buffer,:,:)
						Cx2l(iz+nz-buffer,:,:) = Cx2l(nz-buffer,:,:)
						Cz2l(iz+nz-buffer,:,:) = Cz2l(nz-buffer,:,:)
						Cx1l(iz+nz-buffer,:,:) = Cx1l(nz-buffer,:,:)
						Cz1l(iz+nz-buffer,:,:) = Cz1l(nz-buffer,:,:)
						C0(iz+nz-buffer,:,:) = C0(nz-buffer,:,:)
						Cz1r(iz+nz-buffer,:,:) = Cz1r(nz-buffer,:,:)
						Cx1r(iz+nz-buffer,:,:) = Cx1r(nz-buffer,:,:)
						Cz2r(iz+nz-buffer,:,:) = Cz2r(nz-buffer,:,:)
						Cx2r(iz+nz-buffer,:,:) = Cx2r(nz-buffer,:,:)
						Cz3r(iz+nz-buffer,:,:) = Cz3r(nz-buffer,:,:)
						Cx3r(iz+nz-buffer,:,:) = Cx3r(nz-buffer,:,:)
				enddo
				!$OMP END PARALLEL DO
			
			else
				allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1))	
 				allocate(C0(nz,nx,ny),Cz1r(nz,nx,ny),Cz2r(nz,nx,ny),Cz3r(nz,nx,ny),Cz1l(nz,nx,ny),Cz2l(nz,nx,ny),Cz3l(nz,nx,ny))
 				allocate(Cx1r(nz,nx,ny),Cx2r(nz,nx,ny),Cx3r(nz,nx,ny),Cx1l(nz,nx,ny),Cx2l(nz,nx,ny),Cx3l(nz,nx,ny))
 				C0 = 0.
 				Cz1r = 0.; Cz1l = 0.; Cz2r = 0.; Cz2l = 0.; Cz3r = 0.; Cz3l = 0.;
 				Cx1r = 0.; Cx1l = 0.; Cx2r = 0.; Cx2l = 0.; Cx3r = 0.; Cx3l = 0.;
				!Initializing 6th-order FD div(1/rho grad) 2D
				!$OMP PARALLEL DO private(iz,ix)
				do ix = buffer+1, nx-buffer
					do iz = buffer+1, nz-buffer
						Cx3l(iz,ix,1) = -(1./40.*(b(iz,ix-2,1)+b(iz,ix-1,1)) -11./360.*(b(iz,ix-3,1)+b(iz,ix,1)))*idx
						Cz3l(iz,ix,1) = -(1./40.*(b(iz-2,ix,1)+b(iz-1,ix,1)) -11./360.*(b(iz-3,ix,1)+b(iz,ix,1)))*idz
						Cx2l(iz,ix,1) = -(1./20.*(b(iz,ix-3,1)+b(iz,ix+1,1)) -3./10.*(b(iz,ix-1,1)) +7./40.*(b(iz,ix,1)+b(iz,ix-2,1)))*idx
						Cz2l(iz,ix,1) = -(1./20.*(b(iz-3,ix,1)+b(iz+1,ix,1)) -3./10.*(b(iz-1,ix,1)) +7./40.*(b(iz,ix,1)+b(iz-2,ix,1)))*idz	
						Cx1l(iz,ix,1) = -(-1./40.*(b(iz,ix-3,1)+b(iz,ix+2,1)) -3./10.*(b(iz,ix-2,1)+b(iz,ix+1,1)) -17./40.*(b(iz,ix,1)+b(iz,ix-1,1)))*idx
						Cz1l(iz,ix,1) = -(-1./40.*(b(iz-3,ix,1)+b(iz+2,ix,1)) -3./10.*(b(iz-2,ix,1)+b(iz+1,ix,1)) -17./40.*(b(iz,ix,1)+b(iz-1,ix,1)))*idz
						C0(iz,ix,1) = -((idz+idx)*101.0/180.0*b(iz,ix,1) &
									  +1./180.0*((b(iz-3,ix,1)+b(iz+3,ix,1))*idz+(b(iz,ix-3,1)+b(iz,ix+3,1))*idx) &
									  +1./8.0*((b(iz-2,ix,1)+b(iz+2,ix,1))*idz+(b(iz,ix-2,1)+b(iz,ix+2,1))*idx)   &
									  +19./20.0*((b(iz-1,ix,1)+b(iz+1,ix,1))*idz+(b(iz,ix-1,1)+b(iz,ix+1,1))*idx)) 
						Cz1r(iz,ix,1) = -(-1./40.*(b(iz-2,ix,1)+b(iz+3,ix,1)) -3./10.*(b(iz-1,ix,1)+b(iz+2,ix,1)) -17./40.*(b(iz,ix,1)+b(iz+1,ix,1)))*idz
						Cx1r(iz,ix,1) = -(-1./40.*(b(iz,ix-2,1)+b(iz,ix+3,1)) -3./10.*(b(iz,ix-1,1)+b(iz,ix+2,1)) -17./40.*(b(iz,ix,1)+b(iz,ix+1,1)))*idx
						Cz2r(iz,ix,1) = -(1./20.*(b(iz-1,ix,1)+b(iz+3,ix,1)) -3./10.*(b(iz+1,ix,1)) +7./40.*(b(iz,ix,1)+b(iz+2,ix,1)))*idz	
						Cx2r(iz,ix,1) = -(1./20.*(b(iz,ix-1,1)+b(iz,ix+3,1)) -3./10.*(b(iz,ix+1,1)) +7./40.*(b(iz,ix,1)+b(iz,ix+2,1)))*idx	  
						Cz3r(iz,ix,1) = -(1./40.*(b(iz+1,ix,1)+b(iz+2,ix,1)) -11./360.*(b(iz,ix,1)+b(iz+3,ix,1)))*idz
						Cx3r(iz,ix,1) = -(1./40.*(b(iz,ix+1,1)+b(iz,ix+2,1)) -11./360.*(b(iz,ix,1)+b(iz,ix+3,1)))*idx
					enddo
				enddo
				!$OMP END PARALLEL DO

				
				!Setting the filters in the buffer region
				!$OMP PARALLEL DO private(ix)
				do ix = 1,buffer
						!Extending in the x
						!LEFT x
				 		Cx3l(buffer+1:nz-buffer,ix,1) = Cx3l(buffer+1:nz-buffer,buffer+1,1)
				 		Cz3l(buffer+1:nz-buffer,ix,1) = Cz3l(buffer+1:nz-buffer,buffer+1,1)
				 		Cx2l(buffer+1:nz-buffer,ix,1) = Cx2l(buffer+1:nz-buffer,buffer+1,1)
				 		Cz2l(buffer+1:nz-buffer,ix,1) = Cz2l(buffer+1:nz-buffer,buffer+1,1)
				 		Cx1l(buffer+1:nz-buffer,ix,1) = Cx1l(buffer+1:nz-buffer,buffer+1,1)
				 		Cz1l(buffer+1:nz-buffer,ix,1) = Cz1l(buffer+1:nz-buffer,buffer+1,1)
				 		C0(buffer+1:nz-buffer,ix,1)   = C0(buffer+1:nz-buffer,buffer+1,1)
				 		Cz1r(buffer+1:nz-buffer,ix,1) = Cz1r(buffer+1:nz-buffer,buffer+1,1)
				 		Cx1r(buffer+1:nz-buffer,ix,1) = Cx1r(buffer+1:nz-buffer,buffer+1,1)
				 		Cz2r(buffer+1:nz-buffer,ix,1) = Cz2r(buffer+1:nz-buffer,buffer+1,1)
				 		Cx2r(buffer+1:nz-buffer,ix,1) = Cx2r(buffer+1:nz-buffer,buffer+1,1)
				 		Cz3r(buffer+1:nz-buffer,ix,1) = Cz3r(buffer+1:nz-buffer,buffer+1,1)
				 		Cx3r(buffer+1:nz-buffer,ix,1) = Cx3r(buffer+1:nz-buffer,buffer+1,1)
				 		!RIGHT x
				 		Cx3l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx3l(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz3l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz3l(buffer+1:nz-buffer,nx-buffer,1)
				 		Cx2l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx2l(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz2l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz2l(buffer+1:nz-buffer,nx-buffer,1)
				 		Cx1l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx1l(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz1l(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz1l(buffer+1:nz-buffer,nx-buffer,1)
				 		C0(buffer+1:nz-buffer,ix+nx-buffer,1)   = C0(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz1r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz1r(buffer+1:nz-buffer,nx-buffer,1)
				 		Cx1r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx1r(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz2r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz2r(buffer+1:nz-buffer,nx-buffer,1)
				 		Cx2r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx2r(buffer+1:nz-buffer,nx-buffer,1)
				 		Cz3r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cz3r(buffer+1:nz-buffer,nx-buffer,1)
				 		Cx3r(buffer+1:nz-buffer,ix+nx-buffer,1) = Cx3r(buffer+1:nz-buffer,nx-buffer,1)
				 		!Extending in the z
						!UP z
						iz = ix
						Cx3l(iz,:,1) = Cx3l(buffer+1,:,1)
						Cz3l(iz,:,1) = Cz3l(buffer+1,:,1)
						Cx2l(iz,:,1) = Cx2l(buffer+1,:,1)
						Cz2l(iz,:,1) = Cz2l(buffer+1,:,1)
						Cx1l(iz,:,1) = Cx1l(buffer+1,:,1)
						Cz1l(iz,:,1) = Cz1l(buffer+1,:,1)
						C0(iz,:,1) = C0(buffer+1,:,1)
						Cz1r(iz,:,1) = Cz1r(buffer+1,:,1)
						Cx1r(iz,:,1) = Cx1r(buffer+1,:,1)
						Cz2r(iz,:,1) = Cz2r(buffer+1,:,1)
						Cx2r(iz,:,1) = Cx2r(buffer+1,:,1)
						Cz3r(iz,:,1) = Cz3r(buffer+1,:,1)
						Cx3r(iz,:,1) = Cx3r(buffer+1,:,1)
						!DOWN z
						Cx3l(iz+nz-buffer,:,1) = Cx3l(nz-buffer,:,1)
						Cz3l(iz+nz-buffer,:,1) = Cz3l(nz-buffer,:,1)
						Cx2l(iz+nz-buffer,:,1) = Cx2l(nz-buffer,:,1)
						Cz2l(iz+nz-buffer,:,1) = Cz2l(nz-buffer,:,1)
						Cx1l(iz+nz-buffer,:,1) = Cx1l(nz-buffer,:,1)
						Cz1l(iz+nz-buffer,:,1) = Cz1l(nz-buffer,:,1)
						C0(iz+nz-buffer,:,1) = C0(nz-buffer,:,1)
						Cz1r(iz+nz-buffer,:,1) = Cz1r(nz-buffer,:,1)
						Cx1r(iz+nz-buffer,:,1) = Cx1r(nz-buffer,:,1)
						Cz2r(iz+nz-buffer,:,1) = Cz2r(nz-buffer,:,1)
						Cx2r(iz+nz-buffer,:,1) = Cx2r(nz-buffer,:,1)
						Cz3r(iz+nz-buffer,:,1) = Cz3r(nz-buffer,:,1)
						Cx3r(iz+nz-buffer,:,1) = Cx3r(nz-buffer,:,1)
				enddo
				!$OMP END PARALLEL DO
			endif
			
			!deallocate temporary array for 1/rho
			deallocate(b)
		endif

		!Temporary variable for the Db2 zeroing out
    	tmp = 0.0



		return
	end subroutine Db2_init

  	subroutine Db2_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:) 	 :: mm,dd
    		call Db2_op(adj,add,mm,dd)
  	end subroutine Db2_call

	subroutine Db2_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm !Input 
		real, dimension(nz,nx,ny)			:: dd !Output 
		!Internal variables
		integer						:: it, iz, ix, iy


		tmp(:,:,:) = 0.0 

    	if(adj) then !Adjoint
        	if (.not. add) mm = 0. !Zero-out the model

			if(twoD) then
					!$OMP PARALLEL DO private(iz,ix)
			        do ix=1,nx
            			 do iz=1,nz
                			tmp(iz,ix,1)   = tmp(iz,ix,1)   + C0(iz,ix,1)*dd(iz,ix,1) 
                            tmp(iz+1,ix,1) = tmp(iz+1,ix,1) + Cz1r(iz,ix,1)*dd(iz,ix,1) 
                            tmp(iz-1,ix,1) = tmp(iz-1,ix,1) + Cz1l(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz+2,ix,1) = tmp(iz+2,ix,1) + Cz2r(iz,ix,1)*dd(iz,ix,1)  
                            tmp(iz-2,ix,1) = tmp(iz-2,ix,1) + Cz2l(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz+3,ix,1) = tmp(iz+3,ix,1) + Cz3r(iz,ix,1)*dd(iz,ix,1) 
                            tmp(iz-3,ix,1) = tmp(iz-3,ix,1) + Cz3l(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz,ix+1,1) = tmp(iz,ix+1,1) + Cx1r(iz,ix,1)*dd(iz,ix,1) 
                            tmp(iz,ix-1,1) = tmp(iz,ix-1,1) + Cx1l(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz,ix+2,1) = tmp(iz,ix+2,1) + Cx2r(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz,ix-2,1) = tmp(iz,ix-2,1) + Cx2l(iz,ix,1)*dd(iz,ix,1)
                            tmp(iz,ix+3,1) = tmp(iz,ix+3,1) + Cx3r(iz,ix,1)*dd(iz,ix,1) 
                            tmp(iz,ix-3,1) = tmp(iz,ix-3,1) + Cx3l(iz,ix,1)*dd(iz,ix,1)
                		enddo
            		enddo
            		!$OMP END PARALLEL DO
			
			else

        	 !$OMP PARALLEL DO private(iz,ix,iy)
			 do iy = 1,ny
        		    do ix=1,nx
            			 do iz=1,nz
            			 	tmp(iz,ix,iy)   = tmp(iz,ix,iy)   + C0(iz,ix,iy)*dd(iz,ix,iy) 
                            tmp(iz+1,ix,iy) = tmp(iz+1,ix,iy) + Cz1r(iz,ix,iy)*dd(iz,ix,iy) 
                            tmp(iz-1,ix,iy) = tmp(iz-1,ix,iy) + Cz1l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz+2,ix,iy) = tmp(iz+2,ix,iy) + Cz2r(iz,ix,iy)*dd(iz,ix,iy)  
                            tmp(iz-2,ix,iy) = tmp(iz-2,ix,iy) + Cz2l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz+3,ix,iy) = tmp(iz+3,ix,iy) + Cz3r(iz,ix,iy)*dd(iz,ix,iy) 
                            tmp(iz-3,ix,iy) = tmp(iz-3,ix,iy) + Cz3l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix+1,iy) = tmp(iz,ix+1,iy) + Cx1r(iz,ix,iy)*dd(iz,ix,iy) 
                            tmp(iz,ix-1,iy) = tmp(iz,ix-1,iy) + Cx1l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix+2,iy) = tmp(iz,ix+2,iy) + Cx2r(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix-2,iy) = tmp(iz,ix-2,iy) + Cx2l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix+3,iy) = tmp(iz,ix+3,iy) + Cx3r(iz,ix,iy)*dd(iz,ix,iy) 
                            tmp(iz,ix-3,iy) = tmp(iz,ix-3,iy) + Cx3l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy+1) = tmp(iz,ix,iy+1) +	Cy1r(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy-1) = tmp(iz,ix,iy-1) + Cy1l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy+2) = tmp(iz,ix,iy+2) + Cy2r(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy-2) = tmp(iz,ix,iy-2) + Cy2l(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy+3) = tmp(iz,ix,iy+3) + Cy3r(iz,ix,iy)*dd(iz,ix,iy)
                            tmp(iz,ix,iy-3) = tmp(iz,ix,iy-3) + Cy3l(iz,ix,iy)*dd(iz,ix,iy)
                		enddo
            		enddo
			 enddo
        	 !$OMP END PARALLEL DO
        	 
			endif
			mm(:,:,:)= mm(:,:,:) + tmp(1:nz,1:nx,1:ny)
		
		else	     !Forward
        	if (.not. add) dd = 0. !Zero-out the data
			tmp(1:nz,1:nx,1:ny) = mm 
			if(twoD) then
				!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		dd(iz,ix,1) = dd(iz,ix,1) + C0(iz,ix,1)*tmp(iz,ix,1)     + &
                                  		       Cz1r(iz,ix,1)*tmp(iz+1,ix,1) + Cz1l(iz,ix,1)*tmp(iz-1,ix,1) + &
                                   		       Cz2r(iz,ix,1)*tmp(iz+2,ix,1) + Cz2l(iz,ix,1)*tmp(iz-2,ix,1) + &
                                   		       Cz3r(iz,ix,1)*tmp(iz+3,ix,1) + Cz3l(iz,ix,1)*tmp(iz-3,ix,1) + &
                                   		       Cx1r(iz,ix,1)*tmp(iz,ix+1,1) + Cx1l(iz,ix,1)*tmp(iz,ix-1,1) + &
                                 		       Cx2r(iz,ix,1)*tmp(iz,ix+2,1) + Cx2l(iz,ix,1)*tmp(iz,ix-2,1) + &
                                 		       Cx3r(iz,ix,1)*tmp(iz,ix+3,1) + Cx3l(iz,ix,1)*tmp(iz,ix-3,1)
                                      		       

                	enddo
            	enddo
            	!$OMP END PARALLEL DO
			
			else
  
        	 !$OMP PARALLEL DO private(iz,ix,iy)
			 do iy = 1,ny
        		do ix=1,nx
            		 do iz=1,nz
            			dd(iz,ix,iy) = dd(iz,ix,iy) + C0(iz,ix,iy)*tmp(iz,ix,iy)   + &
                                  		       Cz1r(iz,ix,iy)*tmp(iz+1,ix,iy) + Cz1l(iz,ix,iy)*tmp(iz-1,ix,iy) + &
                                  		       Cz2r(iz,ix,iy)*tmp(iz+2,ix,iy) + Cz2l(iz,ix,iy)*tmp(iz-2,ix,iy) + &
                                  		       Cz3r(iz,ix,iy)*tmp(iz+3,ix,iy) + Cz3l(iz,ix,iy)*tmp(iz-3,ix,iy) + &
                                  		       Cx1r(iz,ix,iy)*tmp(iz,ix+1,iy) + Cx1l(iz,ix,iy)*tmp(iz,ix-1,iy) + &
	                               		       Cx2r(iz,ix,iy)*tmp(iz,ix+2,iy) + Cx2l(iz,ix,iy)*tmp(iz,ix-2,iy) + &
                                   		       Cx3r(iz,ix,iy)*tmp(iz,ix+3,iy) + Cx3l(iz,ix,iy)*tmp(iz,ix-3,iy) + &
					       					   Cy1r(iz,ix,iy)*tmp(iz,ix,iy+1) + Cy1l(iz,ix,iy)*tmp(iz,ix,iy-1) + &
                                  		       Cy2r(iz,ix,iy)*tmp(iz,ix,iy+2) + Cy2l(iz,ix,iy)*tmp(iz,ix,iy-2) + &
                                   		       Cy3r(iz,ix,iy)*tmp(iz,ix,iy+3) + Cy3l(iz,ix,iy)*tmp(iz,ix,iy-3)

            		enddo
           		enddo
			 enddo
        	 !$OMP END PARALLEL DO

			endif
 
		endif

		
		return
	end subroutine Db2_op

	!Subroutine for deallocating variables
	subroutine Db2_close()
		if (allocated(tmp)) then
			deallocate(Cx3l,Cx2l,Cx1l,Cx3r,Cx2r,Cx1r)
			deallocate(Cz3l,Cz2l,Cz1l,Cz3r,Cz2r,Cz1r,C0)
			deallocate(tmp)
		endif
		if (allocated(Cy1r)) then
			deallocate(Cy3l,Cy2l,Cy1l,Cy3r,Cy2r,Cy1r)
		endif
	end subroutine Db2_close

end module Db2
