!Module for computing the div(1/rho grad) with staggered grid density (zero-outside -> Forward=Adjoint)
!8 points stencil of div(1/rho grad)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Output DbD of wavefield
!D(z,x,y) = DbD8pt(Model(z,x,y))
!Model(z,x,y) = DbD8pt*(D(z,x,y))
!It computes the derivatives as a chain of operators:
!Dm*B*Dp


module DbD8pt

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  real, private									:: dz, dx, dy
  real, private									:: C1z, C2z, C3z, C4z,C1x, C2x, C3x, C4x, C1y, C2y, C3y, C4y
  real, private, dimension(:,:,:),allocatable	:: tmp, tmp1, tmp2
  real, private, dimension(:,:,:),allocatable	:: bz, bx, by ! 1/rho
  logical, private								:: twoD

  contains


	!Initialization routine
	subroutine DbD8pt_init(nz_in, nx_in, ny_in, dz_in, dx_in, dy_in, rho_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, intent(in)						:: dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		real, dimension(:,:,:), intent(in)		:: rho_in
		!Internal variables
		integer 								:: buffer, iz, ix, iy

   		parameter(buffer=4)

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dz = dz_in; 
		dx = dx_in; 
		dy = dy_in; 
		twoD = twoD_in

       		if (.not. allocated(tmp)) then
			allocate(bz(nz,nx,ny),bx(nz,nx,ny))
			if (.not. twoD) then
				allocate(by(nz,nx,ny))
				allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer),tmp1(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer),tmp2(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer))	
				!place inverse of density on a staggered grid (Harmonic average Moczo et al. 2007)
				!Bz rho on (i+1/2,j,k)
				!$OMP PARALLEL DO private(iy,ix,iz)
				do iy = 1, ny
					do ix = 1, nx
						do iz = 1, nz-1
							bz(iz,ix,iy) = 0.5*(1./rho_in(iz,ix,iy) + 1./rho_in(iz+1,ix,iy))
						enddo
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				bz(nz,1:nx,1:ny) = 1./rho_in(nz,1:nx,1:ny) !Last depth elements
				!Bx rho on (i,j+1/2,k)
				!$OMP PARALLEL DO private(iy,ix,iz)
				do iy = 1, ny
					do ix = 1, nx-1
						do iz = 1, nz
							bx(iz,ix,iy) = 0.5*(1./rho_in(iz,ix,iy) + 1./rho_in(iz,ix+1,iy))
						enddo
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				bx(1:nz,nx,1:ny) = 1./rho_in(1:nz,nx,1:ny) !Last horizontal-x elements
				!By rho on (i,j,k+1/2)
				!$OMP PARALLEL DO private(iy,ix,iz)
				do iy = 1, ny-1
					do ix = 1, nx
						do iz = 1, nz
							by(iz,ix,iy) = 0.5*(1./rho_in(iz,ix,iy) + 1./rho_in(iz,ix,iy+1))
						enddo
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				by(1:nz,1:nx,ny) = 1./rho_in(1:nz,1:nx,ny) !Last horizontal-y elements

			else
				allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1),tmp1(1-buffer:nz+buffer,1-buffer:nx+buffer,1),tmp2(1-buffer:nz+buffer,1-buffer:nx+buffer,1))
				!place inverse of density on a staggered grid (Harmonic average Moczo et al. 2007)
				!Bz rho on (i+1/2,j,k)
				!$OMP PARALLEL DO private(ix,iz)
				do ix = 1, nx
					do iz = 1, nz-1
						bz(iz,ix,1) = 0.5*(1./rho_in(iz,ix,1) + 1./rho_in(iz+1,ix,1))
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				bz(nz,1:nx,1) = 1./rho_in(nz,1:nx,1) !Last depth elements
				!Bx rho on (i,j+1/2,k)
				!$OMP PARALLEL DO private(ix,iz)
				do ix = 1, nx-1
					do iz = 1, nz
						bx(iz,ix,1) = 0.5*(1./rho_in(iz,ix,1) + 1./rho_in(iz,ix+1,1))
					enddo
				enddo	
				!$OMP END PARALLEL DO
				!Boundary grid points
				bx(1:nz,nx,1) = 1./rho_in(1:nz,nx,1) !Last horizontal-x elements
			endif
		endif
		!Dp coefficients application --> C1(tmp(i+1)-tmp(i))+C2(tmp(i+2)-tmp(i-1))+C3(tmp(i+3)-tmp(i-2))......
		!Dm coefficients application --> C1(tmp(i)-tmp(i-1))+C2(tmp(i+1)-tmp(i-2))+C3(tmp(i+2)-tmp(i-3))......
		!Defining Dm coefficients
		C1z = 1.196289062541883/(dz)
		C2z = -0.079752604188901/(dz)
		C3z = 0.009570312506634/(dz)
		C4z = -6.975446437140719e-04/(dz)
		
		C1x = 1.196289062541883/(dx)
		C2x = -0.079752604188901/(dx)
		C3x = 0.009570312506634/(dx)
		C4x = -6.975446437140719e-04/(dx)

		C1y = 1.196289062541883/(dy)
		C2y = -0.079752604188901/(dy)
		C3y = 0.009570312506634/(dy)
		C4y = -6.975446437140719e-04/(dy)


		!Temporary variable for the DbD zeroing out
    		tmp = 0.0; tmp1 = 0.0; tmp2 = 0.0



		return
	end subroutine DbD8pt_init

  	subroutine DbD8pt_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:) 	 :: mm,dd
    		call DbD8pt_op(adj,add,mm,dd)
  	end subroutine DbD8pt_call

	subroutine DbD8pt_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm !Input 
		real, dimension(nz,nx,ny)			:: dd !Output 
		!Internal variables
		integer						:: it, iz, ix, iy




    	if(adj) then !Adjoint NOTE THAT in staggered grid it is valid Dm*=-Dp and Dp*=-Dm which means (Dm*B*Dp) = (Dp*)B(Dm*)
        	if (.not. add) mm = 0. !Zero-out the model
		tmp(1:nz,1:nx,1:ny) = dd 
			if(twoD) then
				!##############################    (Dzp*)*Bz*(Dzm*)    #############################
				!Applying (Dzm*)
				!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
          				tmp1(iz,ix,1) =  		C1z*(tmp(iz+1,ix,1)-tmp(iz,ix,1))   &
										      + C2z*(tmp(iz+2,ix,1)-tmp(iz-1,ix,1)) &
										      + C3z*(tmp(iz+3,ix,1)-tmp(iz-2,ix,1)) &
										      + C4z*(tmp(iz+4,ix,1)-tmp(iz-3,ix,1)) 
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
				
				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bz(:,:,:)

				!Applying Dzm
        		!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		mm(iz,ix,1) = mm(iz,ix,1)     + C1z*(tmp2(iz,ix,1)-tmp2(iz-1,ix,1))   &
									      + C2z*(tmp2(iz+1,ix,1)-tmp2(iz-2,ix,1)) &
									      + C3z*(tmp2(iz+2,ix,1)-tmp2(iz-3,ix,1)) &
									      + C4z*(tmp2(iz+3,ix,1)-tmp2(iz-4,ix,1))
		            enddo
		     	enddo
		        !$OMP END PARALLEL DO
				!##############################    (Dxp*)*Bx*(Dxm*)    #########################
				!Applying (Dxm*)
				!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		tmp1(iz,ix,1) =  		C1x*(tmp(iz,ix+1,1)-tmp(iz,ix,1))   &
										      + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-1,1)) &
										      + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-2,1)) &
										      + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-3,1))
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
				
				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bx(:,:,:)

				!Applying (Dxp*)
        		!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                				mm(iz,ix,1) = mm(iz,ix,1)     + C1x*(tmp2(iz,ix,1)-tmp2(iz,ix-1,1))   &
									      + C2x*(tmp2(iz,ix+1,1)-tmp2(iz,ix-2,1)) &
									      + C3x*(tmp2(iz,ix+2,1)-tmp2(iz,ix-3,1)) &
									      + C4x*(tmp2(iz,ix+3,1)-tmp2(iz,ix-4,1)) 
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
			else !3D
				!##############################    (Dzp*)*Bz*(Dzm*)    #############################
				!Applying (Dzm*)
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						tmp1(iz,ix,iy) =	      C1z*(tmp(iz,ix,iy)-tmp(iz+1,ix,iy))   &
									   	    				+ C2z*(tmp(iz-1,ix,iy)-tmp(iz+2,ix,iy)) &
									      	    			+ C3z*(tmp(iz-2,ix,iy)-tmp(iz+3,ix,iy)) &
									      	    			+ C4z*(tmp(iz-3,ix,iy)-tmp(iz+4,ix,iy)) 

            					enddo
           				enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bz(:,:,:)

				!Applying (Dzp*)
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						mm(iz,ix,iy) = 	mm(iz,ix,iy)	+ C1z*(tmp2(iz-1,ix,iy)-tmp2(iz,ix,iy))   &
									   		+ C2z*(tmp2(iz-2,ix,iy)-tmp2(iz+1,ix,iy)) &
											+ C3z*(tmp2(iz-3,ix,iy)-tmp2(iz+2,ix,iy)) &
											+ C4z*(tmp2(iz-4,ix,iy)-tmp2(iz+3,ix,iy)) 
            					enddo
           				enddo
			 	enddo
        	 	!$OMP END PARALLEL DO
				!##############################    (Dxp*)*Bx*(Dxm*)    #########################
				!Applying (Dxm*)
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						tmp1(iz,ix,iy) = 		  C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix,iy))   &
									   		+ C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-1,iy)) &
											+ C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-2,iy)) &
											+ C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-3,iy)) 
            					enddo
           				enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bx(:,:,:)

				!Applying (Dxp*)
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						mm(iz,ix,iy) = mm(iz,ix,iy) + C1x*(tmp2(iz,ix,iy)-tmp2(iz,ix-1,iy))   &
									   	   					    + C2x*(tmp2(iz,ix+1,iy)-tmp2(iz,ix-2,iy)) &
									      	    				+ C3x*(tmp2(iz,ix+2,iy)-tmp2(iz,ix-3,iy)) &
									      	    				+ C4x*(tmp2(iz,ix+3,iy)-tmp2(iz,ix-4,iy))
            					enddo
           				enddo
			 	enddo
        	 	!$OMP END PARALLEL DO
        
				!##############################    (Dyp*)*By*(Dym*)    #########################
				!Applying (Dym*)
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				 	do iz=1,nz
       	     				tmp1(iz,ix,iy) =  		  C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy))   &
											   		+ C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-1)) &
													+ C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-2)) &
													+ C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-3)) 
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*by(:,:,:)

				!Applying (Dyp*)
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				    do iz=1,nz
       	     				mm(iz,ix,iy) = mm(iz,ix,iy) + C1y*(tmp2(iz,ix,iy)-tmp2(iz,ix,iy-1))   &
									 			  	    + C2y*(tmp2(iz,ix,iy+1)-tmp2(iz,ix,iy-2)) &
									   			   	    + C3y*(tmp2(iz,ix,iy+2)-tmp2(iz,ix,iy-3)) &
									   			   	    + C4y*(tmp2(iz,ix,iy+3)-tmp2(iz,ix,iy-4)) 
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO
        	

        	 
        	 
			endif
		
		
		else	     !Forward
        	if (.not. add) dd = 0. !Zero-out the data
			tmp(1:nz,1:nx,1:ny) = mm 
			if(twoD) then
				!##############################    Dzm*Bz*Dzp    #############################
				!Applying Dzp
				!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		tmp1(iz,ix,1) =  		C1z*(tmp(iz+1,ix,1)-tmp(iz,ix,1))   &
										      + C2z*(tmp(iz+2,ix,1)-tmp(iz-1,ix,1)) &
										      + C3z*(tmp(iz+3,ix,1)-tmp(iz-2,ix,1)) &
										      + C4z*(tmp(iz+4,ix,1)-tmp(iz-3,ix,1)) 
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
				
				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bz(:,:,:)

				!Applying Dzm
        		!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		dd(iz,ix,1) = dd(iz,ix,1)     + C1z*(tmp2(iz,ix,1)-tmp2(iz-1,ix,1))   &
												      + C2z*(tmp2(iz+1,ix,1)-tmp2(iz-2,ix,1)) &
												      + C3z*(tmp2(iz+2,ix,1)-tmp2(iz-3,ix,1)) &
												      + C4z*(tmp2(iz+3,ix,1)-tmp2(iz-4,ix,1)) 
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
				!##############################    Dxm*Bx*Dxp    #############################
				!Applying Dxp
				!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
        			do iz=1,nz
            				tmp1(iz,ix,1) =  		C1x*(tmp(iz,ix+1,1)-tmp(iz,ix,1))   &
									 		      + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-1,1)) &
									      		  + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-2,1)) &
									      		  + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-3,1))  
		            enddo
		        enddo
		        !$OMP END PARALLEL DO
				
				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bx(:,:,:)

				!Applying Dxm
        		!$OMP PARALLEL DO private(iz,ix)
				do ix=1,nx
            		do iz=1,nz
                		dd(iz,ix,1) = dd(iz,ix,1)     + C1x*(tmp2(iz,ix,1)-tmp2(iz,ix-1,1))   &
									     			 + C2x*(tmp2(iz,ix+1,1)-tmp2(iz,ix-2,1)) &
									 			     + C3x*(tmp2(iz,ix+2,1)-tmp2(iz,ix-3,1)) &
									 			     + C4x*(tmp2(iz,ix+3,1)-tmp2(iz,ix-4,1)) 
		            enddo
		    	enddo
		    	!$OMP END PARALLEL DO
				
			
			else !3D
				!##############################    Dzm*Bz*Dzp    #############################
  				!Applying Dzp
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				 	do iz=1,nz
       	     				tmp1(iz,ix,iy) =  		  C1z*(tmp(iz+1,ix,iy)-tmp(iz,ix,iy))   &
									   				+ C2z*(tmp(iz+2,ix,iy)-tmp(iz-1,ix,iy)) &
													+ C3z*(tmp(iz+3,ix,iy)-tmp(iz-2,ix,iy)) &
													+ C4z*(tmp(iz+4,ix,iy)-tmp(iz-3,ix,iy)) 
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bz(:,:,:)

				!Applying Dzm
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				    do iz=1,nz
       	     				dd(iz,ix,iy) = dd(iz,ix,iy) + C1z*(tmp2(iz,ix,iy)-tmp2(iz-1,ix,iy))   &
									   	   	    		+ C2z*(tmp2(iz+1,ix,iy)-tmp2(iz-2,ix,iy)) &
									    		  	    + C3z*(tmp2(iz+2,ix,iy)-tmp2(iz-3,ix,iy)) &
									      	   			+ C4z*(tmp2(iz+3,ix,iy)-tmp2(iz-4,ix,iy))  
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO
				!##############################    Dxm*Bx*Dxp    #############################
				!Applying Dxp
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				    do iz=1,nz
       	     				tmp1(iz,ix,iy) = 		  C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix,iy))   &
											   		+ C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-1,iy)) &
													+ C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-2,iy)) &
													+ C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-3,iy)) 
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*bx(:,:,:)

				!Applying Dxm
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				 	do iz=1,nz
       	     						dd(iz,ix,iy) = dd(iz,ix,iy) + C1x*(tmp2(iz,ix,iy)-tmp2(iz,ix-1,iy))   &
									   	   					    + C2x*(tmp2(iz,ix+1,iy)-tmp2(iz,ix-2,iy)) &
									      	    				+ C3x*(tmp2(iz,ix+2,iy)-tmp2(iz,ix-3,iy)) &
									      	    				+ C4x*(tmp2(iz,ix+3,iy)-tmp2(iz,ix-4,iy)) 
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO
				!##############################    Dym*By*Dyp    #############################
				!Applying Dyp
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				 	do iz=1,nz
       	     				tmp1(iz,ix,iy) =  		  C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy))   &
											   		+ C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-1)) &
													+ C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-2)) &
													+ C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-3))
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

				!Applying B scaling (1/rho)
				tmp2(1:nz,1:nx,1:ny) = tmp1(1:nz,1:nx,1:ny)*by(:,:,:)

				!Applying Dym
        		!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        			do ix=1,nx
       				    do iz=1,nz
       	     				dd(iz,ix,iy) = dd(iz,ix,iy) + C1y*(tmp2(iz,ix,iy)-tmp2(iz,ix,iy-1))   &
									 			  	    + C2y*(tmp2(iz,ix,iy+1)-tmp2(iz,ix,iy-2)) &
									   			   	    + C3y*(tmp2(iz,ix,iy+2)-tmp2(iz,ix,iy-3)) &
									   			   	    + C4y*(tmp2(iz,ix,iy+3)-tmp2(iz,ix,iy-4))
            			enddo
           			enddo
			 	enddo
        	 	!$OMP END PARALLEL DO

			endif
 
		endif

		
		return
	end subroutine DbD8pt_op

	!Subroutine for deallocating variables
	subroutine DbD8pt_close()
		if (allocated(bz)) deallocate(bz)
		if (allocated(bx)) deallocate(bx)
		if (allocated(by)) deallocate(by)
		if (allocated(tmp)) deallocate(tmp)
	end subroutine DbD8pt_close

end module DbD8pt
