!Module for computing the first derivative operator (see Ali's note) J (First difference in the boundary region)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput wavefield
!D(z,x,y) = G(Model(z,x,y))
!Model(z,x,y) = G*(D(z,x,y))
!NOTE ADJOINT = FORWARD
module G

  implicit none
  !Module variables
  integer, private				:: nz, nx, ny, npad
  real, private					:: dz, dx, dy
  logical, private				:: twoD

  contains


	!Initialization routine
	subroutine G_init(nz_in, nx_in, ny_in, dt_in, dz_in, dx_in, dy_in, nv_pad_in,twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in, nv_pad_in
		real									:: dt_in, dz_in, dx_in, dy_in
		real, dimension(:,:,:), pointer			:: vel_in
		logical, intent(in)						:: twoD_in
		!Internal variables

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dz = dz_in
		dx = dx_in
		dy = dy_in
		npad = nv_pad_in
		twoD = twoD_in



		return
	end subroutine G_init


  	subroutine G_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:)  	 :: mm,dd
    		call G_op(adj,add,mm,dd)
  	end subroutine G_call

	
	!Operator routine
	subroutine G_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables
		integer								:: i


    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model


			!Derivative in the z-axis
			mm(1,:,:) = mm(1,:,:) + dd(1,:,:)/dz
			!$OMP PARALLEL DO private(i)
			do i=2,npad
				mm(i,:,:) = mm(i,:,:) + (dd(i,:,:) - dd(i-1,:,:))/dz
			enddo
			!$OMP END PARALLEL DO
			mm(npad+1,:,:) = mm(npad+1,:,:) - dd(npad,:,:)/dz
			mm(nz-npad,:,:) = mm(nz-npad,:,:) - dd(nz-npad+1,:,:)/dz
			!$OMP PARALLEL DO private(i)
			do i=nz-npad+1,nz-1
				mm(i,:,:) = mm(i,:,:) + (dd(i,:,:) - dd(i+1,:,:))/dz
			enddo
			!$OMP END PARALLEL DO
			mm(nz,:,:) = mm(nz,:,:) + dd(nz,:,:)/dz


			!Derivative in the x-axis
			mm(:,1,:) = mm(:,1,:) + dd(:,1,:)/dx
			!$OMP PARALLEL DO private(i)
			do i=2,npad
				mm(:,i,:) = mm(:,i,:) + (dd(:,i,:) - dd(:,i-1,:))/dx
			enddo
			!$OMP END PARALLEL DO
			mm(:,npad+1,:) = mm(:,npad+1,:) - dd(:,npad,:)/dx
			mm(:,nx-npad,:) = mm(:,nx-npad,:) - dd(:,nx-npad+1,:)/dx
			!$OMP PARALLEL DO private(i)
			do i=nx-npad+1,nx-1
				mm(:,i,:) = mm(:,i,:) + (dd(:,i,:) - dd(:,i+1,:))/dx
			enddo
			!$OMP END PARALLEL DO
			mm(:,nx,:) = mm(:,nx,:) + dd(:,nx,:)/dx

			if(.not. twoD) then
				!Derivative in the y-axis
				mm(:,:,1) = mm(:,:,1) + dd(:,:,1)/dy
				!$OMP PARALLEL DO private(i)
				do i=2,npad
					mm(:,:,i) = mm(:,:,i) + (dd(:,:,i) - dd(:,:,i-1))/dy
				enddo
				!$OMP END PARALLEL DO
				mm(:,:,npad+1) = mm(:,:,npad+1) - dd(:,:,npad)/dy
				mm(:,:,ny-npad) = mm(:,:,ny-npad) - dd(:,:,ny-npad+1)/dy
				!$OMP PARALLEL DO private(i)
				do i=ny-npad+1,ny-1
					mm(:,:,i) = mm(:,:,i) + (dd(:,:,i) - dd(:,:,i+1))/dy
				enddo
				!$OMP END PARALLEL DO
				mm(:,:,ny) = mm(:,:,ny) + dd(:,:,ny)/dy
			endif


		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data

			!Derivative in the z-axis
			!$OMP PARALLEL DO private(i)
			do i=1,npad
				dd(i,:,:) = dd(i,:,:) + (mm(i,:,:) - mm(i+1,:,:))/dz
			enddo
			!$OMP END PARALLEL DO
			!$OMP PARALLEL DO private(i)
			do i=nz-npad+1,nz
				dd(i,:,:) = dd(i,:,:) + (mm(i,:,:) - mm(i-1,:,:))/dz
			enddo
			!$OMP END PARALLEL DO

			!Derivative in the x-axis
			!$OMP PARALLEL DO private(i)
			do i=1,npad
				dd(:,i,:) = dd(:,i,:) + (mm(:,i,:) - mm(:,i+1,:))/dx
			enddo
			!$OMP END PARALLEL DO
			!$OMP PARALLEL DO private(i)
			do i=nx-npad+1,nx
				dd(:,i,:) = dd(:,i,:) + (mm(:,i,:) - mm(:,i-1,:))/dx
			enddo
			!$OMP END PARALLEL DO


			if(.not. twoD) then
				!Derivative in the y-axis
				!$OMP PARALLEL DO private(i)
				do i=1,npad
					dd(:,:,i) = dd(:,:,i) + (mm(:,:,i) - mm(:,:,i+1))/dy
				enddo
				!$OMP END PARALLEL DO
				!$OMP PARALLEL DO private(i)
				do i=ny-npad+1,ny
					dd(:,:,i) = dd(:,:,i) + (mm(:,:,i) - mm(:,:,i-1))/dy
				enddo
				!$OMP END PARALLEL DO
			endif

		endif
		
		return
	end subroutine G_op


end module G
