!!$=head1 NAME
!!$
!!$Geo_uniform - Return the input parameters for propagation using an uniform surface acquisition geometry
!!$
!!$=head1 SYNOPSIS
!!$
!!$Geo_uniform < model.H depth=0. dx= dy= file= >/dev/null
!!$
!!$=head1 DESCRIPTION
!!$
!!$Program for creating an uniform surface acquisition given a velocity model !NOTE = It automatically handles 2D/3D geometry!
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from par
!!$
!!$B<filename> I<file> - File containing x,y,z of the source/recording devices
!!$
!!$B<real> I<depth> - [0.] Depth of the devices
!!$
!!$B<real> I<dx> - [d2] x-axes spacing
!!$
!!$B<real> I<dy> - [d3] y-axes spacing
!!$
!!$
!!$=back
!!$
!!$=head1 SEE ALSO
!!$
!!$Vel, Geometry.sh
!!$
!!$=cut
!%
program Geo_uniform
  
  !Used modules 
  use sep
  
  !Variable declaration
  implicit none
  !Input variables
  character*256  		    			:: file_name
  real									:: dx, dy, depth
  !Internal variables
  real									:: o1, o2, o3, d1, d2, d3
  integer								:: n1, n2, n3, idx, idy
  real									:: x_leng, y_leng
  !Output variables
  integer								:: ndx, ndy !number of devices
  real									:: xpos, ypos
  
  call sep_init (SOURCE)
  
  !reading from input model
  call from_history("n1", n1)
  call from_history("o1", o1);
  call from_history("d1", d1);
  call from_history("n2", n2)
  call from_history("o2", o2);
  call from_history("d2", d2);
  call from_history("n3", n3)
  call from_history("o3", o3);
  call from_history("d3", d3);
  !reading from parameters
  call from_par ("file",file_name)
  call from_par ("dx",dx,d2)
  call from_par ("dy",dy,d3)
  call from_par ("depth",depth,0.)
  
  
  if(n3 .eq. 1) then !2D dataset
  	x_leng = (n2-1) * d2
  	if(x_leng .ge. dx) then
  		ndx = x_leng / dx + 1
  	else
  		write(0,*)"Error x sampling greater than model interval"
  	endif
  	
  	open(file=file_name,unit=24, status='unknown')
  	write(24,*) "# Number of devices"
  	write(24,*) ndx
 	write(24,*) "# Device positions (x,y,z)[m]"
 	do idx = 1, ndx
 		xpos = o2 + (idx-1) * dx
 		write(24,*) xpos, o3, depth
 	enddo
  else !3D dataset
  	x_leng = (n2-1) * d2
  	if(x_leng .ge. dx) then
  		ndx = x_leng / dx + 1
  	else
  		write(0,*)"Error x sampling greater than model interval"
  	endif
  	y_leng = (n3-1) * d3
  	if(y_leng .ge. dy) then
  		ndy = y_leng / dy + 1
  	else
  		write(0,*)"Error y sampling greater than model interval"
  	endif
  	
  	open(file=file_name,unit=24, status='unknown')
  	write(24,*) "# Number of devices"
  	write(24,*) ndx*ndy
 	write(24,*) "# Device positions (x,y,z)[m]"
 	do idx = 1, ndx
 		xpos = o2 + (idx-1) * dx
 		do idy = 1, ndy
 			ypos = o3 + (idy-1) * dy
 			write(24,*) xpos, ypos, depth
 		enddo
 	enddo
  	
  endif
  
  
  close(24)
  call sep_close()
  call exit (0)
end program Geo_uniform
