!!$=head1 NAME
!!$
!!$WEMVA-like component of the FULL FWI HESSIAN - Acoustic Hessian Wemva-component operator with constant density.
!!$
!!$=head1 SYNOPSIS
!!$
!!$Hessian_2nd_adj.x <velocity.H deltadata=data.H image=image.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefields=wavefields.H wavefieldr=wavefieldr.H >deltamodel.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation WEMVA-like component of Full FWI Hessian for acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model, the perturbation in the data and returns the perturbed model given an image. !NOTE= it assumes that we can keep in memory the source and receiver wavefields at source-wavelet sampling!
!!$It is the adjoint operator of the second-order scattering operator.
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<velocity.H> - Velocity model in SEP3D format (theoretically [m/s]). (Constructed using Velpar and/or Vel)
!!$
!!$=item from par
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<deltadata> - Perturbed data.
!!$
!!$B<filename> I<image> - Model perturbation to apply the operator.
!!$
!!$B<integer> I<sqslow> - [0] If 1 model perturbation is slowness square; otherwise, velocity perturbation is assumed.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefields.
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save (wantwavefield=1) or read (readwavefield=1) source wavefield
!!$
!!$B<filename> I<wavefieldr> - [wavefieldr.H] File where to save (wantwavefield=1) or read (readwavefield=1) receiver wavefield
!!$
!!$B<filename> I<scatfields> - [scatfields.H] File where to save (wantwavefield=1) or read (readwavefield=1) the scattered source  wavefield
!!$
!!$B<filename> I<scatfieldr> - [scatfieldr.H] File where to save the scattered adj wavefield source side if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<padded> - [0] The model perturbation was already padded (100 elements)
!!$
!!$B<integer> I<readwavefield> - [0] If 1 the program reads the source and receiver wavefields from file.
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<vel_padded> - [0] Velocity model was padded (100 elements), useful for FWI
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Hessian_wemva
  
  !Used modules 
  use sep
  use read_pos
  use sec_deriv
  use Pad_vel
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_step
  use weight_sec
  use Pad_pert
  use sec_deriv_field


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: vel   !velocity model 
  real, dimension(:,:,:), allocatable					:: model, image  !velocity perturbation, and image 
  real, dimension(:), allocatable						:: wavelet, wavelet_buff !Source wavelet 
  integer												:: wantwavefield, readwavefield, verbose
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right
  real, dimension(:,:,:), allocatable					:: sec_scale, field_tmp
  real, dimension(:,:,:,:), allocatable					:: source_wavefield, rec_wavefield
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold, ptemp
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: vel_pad, image_pad  !Padded velocity model and image
  real, dimension(:,:,:), pointer						:: scat_sou_right !for imaging condition
  real													:: dt !Temporal sampling of the propagator
  real, dimension(:), allocatable						:: wavelet_fine, scaled_wavelet !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it, it2, itw !Cycles variables
  integer												:: iz, ix, iy !#######Cycles variables for fixing problem with ifort########
  integer												:: wc, srite, rc, sreed, sk, sseek
  integer												:: desample, padded, vel_padded, sqslow
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr, total_time
  integer												:: count1, count2,ierr, icount, num_threads
  real*8, dimension(:,:), allocatable					:: interp_filter
  integer*8												:: rbit !necessary for reading big files
  !Variable for receivers and source scattered wavefield
  real, dimension(:,:,:), pointer						:: pscattered_left, pscattered_right
  real, dimension(:,:,:), allocatable,target			:: scattered_left, scattered_right
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  call from_par ("verbose",verbose,0)
		  

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)"####################Acoustic 2D/3D wave propagator#####################"
  if(verbose)write(0,*)"####################  WEMVA-like Full FWI Hessian  ####################"  
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("readwavefield",readwavefield,0)
  call from_par ("vel_padded",vel_padded,0)
  call from_par ("desample",desample,1)
  call from_par ("dt",dt,0.)
  call from_par ("padded",padded,0)
  call from_par ("sqslow",sqslow,0)
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the velocity
  if(vel_padded .eq. 0) then
  	allocate(vel(nz,nx,ny))
  	vel(:,:,:) = 0.
  	rc=sreed("in" , vel , nz*nx*ny*4)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	allocate(vel_pad(nz,nx,ny))
  	vel_pad(:,:,:) = 0.
  	rc=sreed("in" , vel_pad, nz*nx*ny*4)
  endif
  if(vel_padded .eq. 1 .and. padded .eq. 0) then
  	write(0,*)"ERROR WHEN VELOCITY PADDED, PERTURBATION MUST BE PADDED TOO"
  	stop
  endif
  !Adding absorbing boundaries to the velocity model and perturbation
  if (padded .eq. 0) then
  	!Not padded
  	allocate(image(nz,nx,ny));image(:,:,:) = 0.
  	rc=sreed("image" , image , nz*nx*ny*4)
  	call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	call Pad_perturbation(image,image_pad,nv_pad)
  else
  	!Padded
  	if(vel_padded .eq. 0) call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	allocate(image_pad(nz,nx,ny));image_pad(:,:,:) = 0.
  	rc=sreed("image" , image_pad , nz*nx*ny*4)
  endif  
  
  !Allocating perturbation output of Wemva-like operator
  allocate(model(nz,nx,ny))
  model(:,:,:) = 0.
  !compute model perturbation from velocity perturbation
  if(sqslow .eq. 0) image_pad(:,:,:) = -2.0*image_pad(:,:,:)/(vel_pad(:,:,:)*vel_pad(:,:,:)*vel_pad(:,:,:))
  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minval(vel_pad)/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxval(vel_pad)/min(dx,dy,dz)
  if(verbose)write(0,*) ""


  !Computing the second order derivative on the source wavelet
  !allocate(wavelet_buff(nw))
  !call sec_deriv_init(nw, dw)
  call sec_deriv_field_init(nw, dw)
  !call sec_deriv_op(.false., .false., wavelet, wavelet_buff)
  !wavelet = wavelet_buff

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_step_init(nz, nx, ny, nt, nv_pad, vel_pad, dt, dz, dx, dy,twoD)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,vel_pad,twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  allocate(scat_sou_right(nz,nx,ny),sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = -dt*dt*vel_pad*vel_pad
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !allocating memory for source wavefield
  allocate(source_wavefield(nz,nx,ny,nw),rec_wavefield(nz,nx,ny,nw))
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))
  allocate(field_tmp(nz,nx,ny), scattered_left(nz,nx,ny), scattered_right(nz,nx,ny))
  pscattered_left=>scattered_left
  pscattered_right=>scattered_right


  
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  
  if (wantwavefield) then
	if(verbose)write(0,*)"#Program will be writing wavefield as requested#"
	if(.not. readwavefield) then
		do is = 1,ns
			do it = 1, nw
				wc=srite("wavefieldr",pcur,nz*nx*ny*4)
			enddo
		enddo
		sk = sseek("wavefieldr",0,0)
	endif
	do is = 1,ns
		do it = 1, nw
			wc=srite("scatfieldr",pcur,nz*nx*ny*4)
		enddo
	enddo
	sk = sseek("scatfieldr",0,0)
  endif
  
  total_time = 0.
  rbit = nz*nx*ny
  rbit = rbit*nw*4
  do is = 1,ns
 	!moving the file writing position with sseek
  	if (wantwavefield) sk = sseek("scatfieldr",rbit,1)
  	old = 0.; cur = 0.
  	source_wavefield = 0.; rec_wavefield = 0.
  	sec_source_left=0.; sec_source_right=0.
  	scat_sou_right = 0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
  	!zero-out the model
	model(:,:,:) = 0.
	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	call system_clock(count_rate=cr)
	call system_clock(count1)
	
!##################################################################	

	if(.not. readwavefield) then
		if (wantwavefield) sk = sseek("wavefieldr",rbit,1)
		!Scaling the source
		scaled_wavelet = wavelet_fine
		call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
		call Sou_scale_op(scaled_wavelet,vel_pad) !Adjoint = Forward (input-output same variable)
		if(verbose)write(0,*)"Propagating source wavefield..."
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
		endif
		do itw = 1, nw-1 !Forward
			it = (itw-1)*desample
			!Propagate source field	
			do it2 = 1, desample
				call Prop_step_op(.false., pold, pcur, pnew)
				!Adding source (perturbed and source wavefield)
				if (twoD) then
					call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				else
					call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				endif	
				!Apply adjoint(fine to coarse) of linear interpolation
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,itw) = source_wavefield(iz,ix,iy,itw) + pcur(iz,ix,iy)*interp_filter(it2,1)
							source_wavefield(iz,ix,iy,itw+1) = source_wavefield(iz,ix,iy,itw+1) + pcur(iz,ix,iy)*interp_filter(it2,2)	
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
				!Swapping pointers
				pold => pcur 
				pcur => pnew
				pnew => pold
			enddo
			!saving source wavefield
			if (wantwavefield) then
				wc=srite("wavefields",source_wavefield(:,:,:,itw),nz*nx*ny*4)
			endif
		
			if (mod(itw,(nw/10))==0) then
        		if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        		close(0)
      		endif
      	
		enddo
		!Last step of the source wavefield just saving source wavefield
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,nw) = source_wavefield(iz,ix,iy,nw) + pcur(iz,ix,iy)*interp_filter(1,1)
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,nw),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
		if(verbose)write(0,'(a)',advance="yes")""
		if(verbose)write(0,*)"End Propagation source wavefield"
	else	
		if(verbose)write(0,*)"Reading source wavefield"
		rc=sreed("wavefields",source_wavefield,rbit)
	endif
	
!##################################################################	
	
	!Receiver wavefield 
	if(.not. readwavefield) then
		if(verbose)write(0,*)"Propagating receiver wavefield..."
		icount = 0.
		old = 0.; cur = 0.
		scat_sou_right = 0.
		poldpert => old
  		pcurpert => cur
  		pnewpert => old
		!Creating "adjoint" source at the receivers
		if (dw .le. dt) then
			do ig = 1,ng
				rc=sreed("deltadata",recfield,nw*4)
				recfield_fine(:,ig) = recfield 
			enddo
		else
			do ig = 1,ng
				rc=sreed("deltadata",recfield,nw*4)
				call resample_op(.false.,.false.,recfield,recfield_fine(:,ig))
			enddo
		endif
		do itw = nw, 2, -1 !Adjoint 
			!Propagate scattered wavefield
			it = (itw-2)*desample
			do it2 = desample+1, 2, -1
				call Prop_step_op(.true., poldpert, pcurpert, pnewpert)
				!Adding receiver deltadata 
				if (twoD) then
					do ig = 1, ng
						call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
						call Add_source2D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
					enddo
				else
					do ig = 1,ng
						call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
						call Add_source3D_op(.true.,.true.,pnewpert,recfield_fine(it+it2,ig))
					enddo
				endif
				!Apply adjoint(fine to coarse) of linear interpolation 
				!####ADDED LOOPS BECAUSE PROBLEM OF IFORT ON SEP MACHINE######
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							psec_source_left(iz,ix,iy) = psec_source_left(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(it2,1)
							psec_source_right(iz,ix,iy) = psec_source_right(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(it2,2)
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
				
				
				!Swapping pointers
				poldpert => pcurpert 
				pcurpert => pnewpert
				pnewpert => poldpert
			enddo
		
			!Scale secondary source (i.e., back-propagated data)
			scat_sou_right(:,:,:)=psec_source_right(:,:,:)*sec_scale(:,:,:)
			call weight_sec_op(.true., .false., rec_wavefield(:,:,:,itw), scat_sou_right(:,:,:))
			!call weight_sec_op(.true., .false., rec_wavefield(:,:,:,itw), psec_source_right(:,:,:)*sec_scale(:,:,:))
		
			!saving source wavefield
			if (wantwavefield) then
				sk = sseek("wavefieldr",-4*nz*nx*ny,1)
				wc=srite("wavefieldr",rec_wavefield(:,:,:,itw),nz*nx*ny*4)
				sk = sseek("wavefieldr",-4*nz*nx*ny,1)
			endif
			icount = icount + 1
			if (mod(icount,(nw/10))==0) then
	   		     	if(verbose)write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
	   		     	close(0)
			endif	
		
			!Swap secondary source left and right pointers
			ptemp => psec_source_left
			psec_source_left => psec_source_right
			psec_source_right => ptemp
			psec_source_left = 0.
		enddo
			!Last step (already propagated)
			itw = 1
			!####ADDED LOOPS BECAUSE PROBLEM OF IFORT ON SEP MACHINE######
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						psec_source_right(iz,ix,iy) = psec_source_right(iz,ix,iy) + pcurpert(iz,ix,iy)*interp_filter(1,2)
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
			!Scale secondary source (i.e., back-propagated data)
			scat_sou_right(:,:,:)=psec_source_right(:,:,:)*sec_scale(:,:,:)
			call weight_sec_op(.true., .false., rec_wavefield(:,:,:,itw), scat_sou_right(:,:,:))
			!call weight_sec_op(.true., .false., rec_wavefield(:,:,:,itw), psec_source_right(:,:,:)*sec_scale(:,:,:))
			!saving source wavefield
			if (wantwavefield) then
				sk = sseek("wavefieldr",-4*nz*nx*ny,1)
				wc=srite("wavefieldr",rec_wavefield(:,:,:,itw),nz*nx*ny*4)
				sk = sseek("wavefieldr",-4*nz*nx*ny,1)
			endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
       	close(0)
		if(verbose)write(0,'(a)',advance="yes")""
		if(verbose)write(0,*)"End Propagation receiver wavefield"
		!moving the file writing position with sseek
  		if (wantwavefield) sk = sseek("wavefieldr",rbit,1)
  	else
  		if(verbose)write(0,*)"Reading receiver wavefield"
		rc=sreed("wavefieldr",rec_wavefield,rbit)
  	endif
	
!##################################################################	

	!Scattered receiver wavefield and imaging condition (WEMVA_s source side)
	if(verbose)write(0,*)"Propagating scattered receiver wavefield..."
	icount = 0.
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
  	sec_source_left=0.; sec_source_right=0.
  	scat_sou_right = 0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
  	
  	!Creating scattered receiver wavefield with image
	!right sample receiver wavefield
	call sec_deriv_field_op(rec_wavefield, field_tmp, nw)
	pscattered_right(:,:,:) = field_tmp(:,:,:) * image_pad(:,:,:)
	
	do itw = nw, 2, -1 !Adjoint 
		!Propagate scattered wavefield
		it = (itw-2)*desample
		
		!Creating scattered receiver wavefield with image
		!left sample receiver wavefield
		call sec_deriv_field_op(rec_wavefield, field_tmp, itw-1)
		pscattered_left(:,:,:) = field_tmp(:,:,:) * image_pad(:,:,:)

		do it2 = desample+1, 2, -1
			call Prop_step_op(.true., poldpert, pcurpert, pnewpert)
			!Adding scattered back-propagated data source
			pnewpert(:,:,:) = pnewpert(:,:,:) + pscattered_left(:,:,:)*interp_filter(it2,1) + pscattered_right(:,:,:)*interp_filter(it2,2)
			!Apply adjoint(fine to coarse) of linear interpolation
			!if(verbose)write(0,*)"adjoint pert it",it+it2
			psec_source_left = psec_source_left + pcurpert*interp_filter(it2,1)
			psec_source_right = psec_source_right + pcurpert*interp_filter(it2,2)
			
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		
		!Scale secondary source (i.e., back-propagated data)
		call weight_sec_op(.true., .false., scat_sou_right(:,:,:), psec_source_right(:,:,:)*sec_scale(:,:,:))
		!Apply imaging condition 
		call sec_deriv_field_op(source_wavefield, field_tmp, itw)
		model(:,:,:) = model(:,:,:) + field_tmp(:,:,:) * scat_sou_right(:,:,:)
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek("scatfieldr",-4*nz*nx*ny,1)
			wc=srite("scatfieldr",scat_sou_right(:,:,:),nz*nx*ny*4)
			sk = sseek("scatfieldr",-4*nz*nx*ny,1)
		endif
		icount = icount + 1
		if (mod(icount,(nw/10))==0) then
	        	if(verbose)write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
	        	close(0)
		endif	
		
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_left = 0.
		
		ptemp => pscattered_left
		pscattered_left => pscattered_right
		pscattered_right => ptemp
		pscattered_left = 0.
	enddo
	!Last step (already propagated)
	itw = 1
	psec_source_right = psec_source_right + pcurpert*interp_filter(desample+1,2)
	!Scale secondary source (i.e., back-propagated data)
		call weight_sec_op(.true., .false., scat_sou_right(:,:,:), psec_source_right(:,:,:)*sec_scale(:,:,:))
		!Apply imaging condition 
		call sec_deriv_field_op(source_wavefield, field_tmp, itw)
		model(:,:,:) = model(:,:,:) + field_tmp(:,:,:) * scat_sou_right(:,:,:)
		!saving source wavefield
		if (wantwavefield) then
			sk = sseek("scatfieldr",-4*nz*nx*ny,1)
			wc=srite("scatfieldr",scat_sou_right(:,:,:),nz*nx*ny*4)
			sk = sseek("scatfieldr",-4*nz*nx*ny,1)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	if(verbose)write(0,*)"End Propagation receiver scattered wavefield"
   	!moving the file writing position with sseek
  	if (wantwavefield) sk = sseek("scatfieldr",rbit,1)

	
!##################################################################	
	!Scattered source wavefield and imaging condition (WEMVA_s receiver side)
	if (.not. readwavefield) then
		if(verbose)write(0,*)"Propagating scattered source wavefield..."
		old = 0.; cur = 0.
		poldpert => old
  		pcurpert => cur
  		pnewpert => old
  		sec_source_left=0.; sec_source_right=0.
  		psec_source_left=>sec_source_left
  		psec_source_right=>sec_source_right
  		scattered_left = 0.
  		scattered_right = 0.
  		pscattered_left=>scattered_left
  		pscattered_right=>scattered_right
  		!Apply the adjoint of imaging condition (i.e., Born operator)
		!left sample secondary source
		field_tmp = 0.
		call sec_deriv_field_op(source_wavefield, field_tmp, 1)
		field_tmp(:,:,:) = field_tmp(:,:,:) * image_pad(:,:,:) * sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., field_tmp(:,:,:), pscattered_left(:,:,:))
  	else
  		source_wavefield = 0.
  		if(verbose)write(0,*)"Reading scattered source wavefield"
		rc=sreed("scatfields",source_wavefield,rbit)
		if(verbose)write(0,*)"Imaging scattered source wavefield with receiver wavefield"
  	endif
  	
	do itw = 1, nw-1 !Forward
		if(.not. readwavefield) then
			it = (itw-1)*desample
			!Apply the adjoint of imaging condition (i.e., Born operator)
			!right sample secondary source
			field_tmp = 0.
			call sec_deriv_field_op(source_wavefield, field_tmp, itw+1)
			field_tmp(:,:,:) = field_tmp(:,:,:) * image_pad(:,:,:) * sec_scale(:,:,:)
			!Scale secondary source
			call weight_sec_op(.false., .false., field_tmp(:,:,:), pscattered_right(:,:,:))
		
			!Propagate scattered wavefield
			do it2 = 1, desample
				call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
				!Adding source (perturbed and source wavefield)
				pnewpert(:,:,:) = pnewpert(:,:,:) + pscattered_left(:,:,:)*interp_filter(it2,1) + pscattered_right(:,:,:)*interp_filter(it2,2)

				!Apply adjoint(fine to coarse) of linear interpolation
				psec_source_left = psec_source_left + pcurpert*interp_filter(it2,1)
				psec_source_right = psec_source_right + pcurpert*interp_filter(it2,2)
				!Swapping pointers
				poldpert => pcurpert 
				pcurpert => pnewpert
				pnewpert => poldpert
			enddo
			!saving scattered wavefield
			if (wantwavefield) then
				wc=srite("scatfields",psec_source_left,nz*nx*ny*4)
			endif
			
		else
			psec_source_left = source_wavefield(:,:,:,itw)
		endif
		call sec_deriv_field_op(rec_wavefield, field_tmp, itw)
		!Apply imaging condition 
		model(:,:,:) = model(:,:,:) + psec_source_left(:,:,:) * field_tmp(:,:,:)
		
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
      	!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_right = 0.
			
		ptemp => pscattered_left
		pscattered_left => pscattered_right
		pscattered_right => ptemp
		pscattered_right = 0.
      	
	enddo
		itw = nw
		if(.not. readwavefield) then
			psec_source_left = psec_source_left + pcurpert*interp_filter(1,1)
			!saving last scattered wavefield step
			if (wantwavefield) then
				wc=srite("scatfields",psec_source_left,nz*nx*ny*4)
			endif
		else
			psec_source_left = source_wavefield(:,:,:,itw)
		endif
			!Apply imaging condition 
			call sec_deriv_field_op(rec_wavefield, field_tmp, itw)
			model(:,:,:) = model(:,:,:) + psec_source_left(:,:,:) * field_tmp(:,:,:)
			if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
			close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	if(verbose)write(0,*)"End Propagation/Imaging source scattered wavefield"
	
!##################################################################	
	
	
	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	total_time = total_time + (count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing model for source n°: ", is
	!Write model for is source
	!compute velocity perturbation from model perturbation
  	if(sqslow .eq. 0) model(:,:,:) = -2.0*model(:,:,:)/(vel_pad(:,:,:)*vel_pad(:,:,:)*vel_pad(:,:,:))
	wc=srite("out",model(:,:,:),nz*nx*ny*4)
   	if(verbose)write(0,*) ""

  enddo
  
  if(verbose)write(0,*) "Total propagation time: ", total_time/60.0 ," minutes"
  
  !Closing used routines
  call Prop_step_close()
  call weight_sec_close()

  !Writing information to headers
  call to_history("n1",nz)
  call to_history("o1",oz)
  call to_history("d1",dz)
  call to_history("label1","z [m]")
  call to_history("n2",nx)
  call to_history("o2",ox)
  call to_history("d2",dx)
  call to_history("label2","x [m]")
  call to_history("n3",ny)
  call to_history("o3",oy)
  call to_history("d3",dy)
  call to_history("label3","y [m]")
  call to_history("n4",ns)
  call to_history("o4",1.)
  call to_history("d4",1.)
  call to_history("label4","Source number")
  
  
  if (wantwavefield) then
  
  	if(twoD) then
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
  			call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",nt/desample+1,"wavefields")
  			call to_history("o3",ow,"wavefields")
 		 	call to_history("d3",dt*desample,"wavefields")
  			call to_history("label3","Time [s]","wavefields")
  			call to_history("n4",ns,"wavefields")
  			call to_history("o4",1.,"wavefields")
  			call to_history("d4",1.,"wavefields")
  			call to_history("label4","Source number","wavefields") 
  		
  		
  			call to_history("n1",nz,"wavefieldr")
  			call to_history("o1",oz,"wavefieldr")
  			call to_history("d1",dz,"wavefieldr")
  			call to_history("label1","z [m]","wavefieldr")
			call to_history("n2",nx,"wavefieldr")
		  	call to_history("o2",ox,"wavefieldr")
  			call to_history("d2",dx,"wavefieldr")
  			call to_history("label2","x [m]","wavefieldr")
			call to_history("n3",nt/desample+1,"wavefieldr")
  			call to_history("o3",ow,"wavefieldr")
 		 	call to_history("d3",dt*desample,"wavefieldr")
  			call to_history("label3","Time [s]","wavefieldr")
  			call to_history("n4",ns,"wavefieldr")
  			call to_history("o4",1.,"wavefieldr")
  			call to_history("d4",1.,"wavefieldr")
  			call to_history("label4","Source number","wavefieldr") 
  		endif 
  			
  			call to_history("n1",nz,"scatfields")
  			call to_history("o1",oz,"scatfields")
  			call to_history("d1",dz,"scatfields")
  			call to_history("label1","z [m]","scatfields")
			call to_history("n2",nx,"scatfields")
		  	call to_history("o2",ox,"scatfields")
  			call to_history("d2",dx,"scatfields")
  			call to_history("label2","x [m]","scatfields")
			call to_history("n3",nt/desample+1,"scatfields")
  			call to_history("o3",ow,"scatfields")
 		 	call to_history("d3",dt*desample,"scatfields")
  			call to_history("label3","Time [s]","scatfields")
  			call to_history("n4",ns,"scatfields")
  			call to_history("o4",1.,"scatfields")
  			call to_history("d4",1.,"scatfields")
  			call to_history("label4","Source number","scatfields") 
  		
  		
  			call to_history("n1",nz,"scatfieldr")
  			call to_history("o1",oz,"scatfieldr")
  			call to_history("d1",dz,"scatfieldr")
  			call to_history("label1","z [m]","scatfieldr")
			call to_history("n2",nx,"scatfieldr")
		  	call to_history("o2",ox,"scatfieldr")
  			call to_history("d2",dx,"scatfieldr")
  			call to_history("label2","x [m]","scatfieldr")
			call to_history("n3",nt/desample+1,"scatfieldr")
  			call to_history("o3",ow,"scatfieldr")
 		 	call to_history("d3",dt*desample,"scatfieldr")
  			call to_history("label3","Time [s]","scatfieldr")
  			call to_history("n4",ns,"scatfieldr")
  			call to_history("o4",1.,"scatfieldr")
  			call to_history("d4",1.,"scatfieldr")
  			call to_history("label4","Source number","scatfieldr") 		
  	else
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
	  		call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",ny,"wavefields")
		  	call to_history("o3",oy,"wavefields")
		  	call to_history("d3",dy,"wavefields")
  			call to_history("label3","y [m]","wavefields")
			call to_history("n4",nt/desample+1,"wavefields")
  			call to_history("o4",ow,"wavefields")
 		 	call to_history("d4",dt*desample,"wavefields")
  			call to_history("label4","Time [s]","wavefields")
  			call to_history("n5",ns,"wavefields")
  			call to_history("o5",1.,"wavefields")
  			call to_history("d5",1.,"wavefields")
  			call to_history("label5","Source number","wavefields")
  		
  		
  			call to_history("n1",nz,"wavefieldr")
  			call to_history("o1",oz,"wavefieldr")
  			call to_history("d1",dz,"wavefieldr")
  			call to_history("label1","z [m]","wavefieldr")
			call to_history("n2",nx,"wavefieldr")
		  	call to_history("o2",ox,"wavefieldr")
  			call to_history("d2",dx,"wavefieldr")
  			call to_history("label2","x [m]","wavefieldr")
			call to_history("n3",ny,"wavefieldr")
		  	call to_history("o3",oy,"wavefieldr")
		  	call to_history("d3",dy,"wavefieldr")
  			call to_history("label3","y [m]","wavefieldr")
			call to_history("n4",nt/desample+1,"wavefieldr")
  			call to_history("o4",ow,"wavefieldr")
 	 		call to_history("d4",dt*desample,"wavefieldr")
  			call to_history("label4","Time [s]","wavefieldr")
  			call to_history("n5",ns,"wavefieldr")
  			call to_history("o5",1.,"wavefieldr")
  			call to_history("d5",1.,"wavefieldr")
  			call to_history("label5","Source number","wavefieldr")
  		endif
  			
  			call to_history("n1",nz,"scatfields")
  			call to_history("o1",oz,"scatfields")
  			call to_history("d1",dz,"scatfields")
  			call to_history("label1","z [m]","scatfields")
			call to_history("n2",nx,"scatfields")
		  	call to_history("o2",ox,"scatfields")
  			call to_history("d2",dx,"scatfields")
  			call to_history("label2","x [m]","scatfields")
			call to_history("n3",ny,"scatfields")
		  	call to_history("o3",oy,"scatfields")
		  	call to_history("d3",dy,"scatfields")
  			call to_history("label3","y [m]","scatfields")
			call to_history("n4",nt/desample+1,"scatfields")
  			call to_history("o4",ow,"scatfields")
 	 		call to_history("d4",dt*desample,"scatfields")
  			call to_history("label4","Time [s]","scatfields")
  			call to_history("n5",ns,"scatfields")
  			call to_history("o5",1.,"scatfields")
  			call to_history("d5",1.,"scatfields")
  			call to_history("label5","Source number","scatfields")
  		
  		
  			call to_history("n1",nz,"scatfieldr")
  			call to_history("o1",oz,"scatfieldr")
  			call to_history("d1",dz,"scatfieldr")
  			call to_history("label1","z [m]","scatfieldr")
			call to_history("n2",nx,"scatfieldr")
		  	call to_history("o2",ox,"scatfieldr")
  			call to_history("d2",dx,"scatfieldr")
  			call to_history("label2","x [m]","scatfieldr")
			call to_history("n3",ny,"scatfieldr")
		  	call to_history("o3",oy,"scatfieldr")
		  	call to_history("d3",dy,"scatfieldr")
  			call to_history("label3","y [m]","scatfieldr")
			call to_history("n4",nt/desample+1,"scatfieldr")
  			call to_history("o4",ow,"scatfieldr")
 	 		call to_history("d4",dt*desample,"scatfieldr")
  			call to_history("label4","Time [s]","scatfieldr")
  			call to_history("n5",ns,"scatfieldr")
  			call to_history("o5",1.,"scatfieldr")
  			call to_history("d5",1.,"scatfieldr")
  			call to_history("label5","Source number","scatfieldr")	
  	endif
  endif
  if(verbose)write(0,*) "End propagation"

  !Deallocating memory
  deallocate(scattered_right,scattered_left,field_tmp)
  deallocate(interp_filter,rec_wavefield,source_wavefield)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale,scat_sou_right)
  deallocate(cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(model,image_pad,vel_pad)
  deallocate(wavelet)
  !deallocate(wavelet_buff,wavelet)
  call sep_close()
  call exit (0)
end program Hessian_wemva

