!!$=head1 NAME
!!$
!!$Tomo-like component of the FULL FWI HESSIAN - Acoustic tomography operator with constant density (dm->dm->ddata).
!!$
!!$=head1 SYNOPSIS
!!$
!!$Hessian_2nd_fw.x <velocity.H image2=image2.H image1=image1.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p >deltadata.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation tomography operator for acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model, scatters it twice, and returns the perturbed data. !NOTE= it assumes that we can keep in memory the source and scattered wavefields at source-wavelet sampling!
!!$It is the forward operator of the second-order scattering operator.
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<velocity.H> - Velocity model in SEP3D format (theoretically [m/s]). (Constructed using Velpar and/or Vel)
!!$
!!$=item from par
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<image1> - First model perturbation to apply the operator.
!!$
!!$B<filename> I<image2> - Second model perturbation to apply the operator.
!!$
!!$B<integer> I<sqslow> - [0] If 1 model perturbation is slowness square; otherwise, velocity perturbation is assumed.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefields.
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save (wantwavefield=1) or read (readwavefield=1) source wavefield
!!$
!!$B<filename> I<scatfield1> - [scatfield1.H] File where to save the scattered source wavefield off of image1 if wantwavefield=1
!!$
!!$B<filename> I<scatfield2> - [scatfield2.H] File where to save the scattered source wavefield off of image2 if wantwavefield=1
!!$
!!$B<filename> I<scatfield12> - [scatfield12.H] File where to save the double-scattered wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<padded> - [0] The model perturbation was already padded (100 elements)
!!$
!!$B<integer> I<readwavefield> - [0] If 1 the program reads the source wavefield from file.
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<vel_padded> - [0] Velocity model was padded (100 elements), useful for FWI
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Hessian_tomo
  
  !Used modules 
  use sep
  use read_pos
  use sec_deriv
  use Pad_vel
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_step
  use weight_sec
  use Pad_pert
  use sec_deriv_field


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: vel   !velocity model 
  real, dimension(:,:,:), allocatable					:: image1, image2  !image1, and image2 
  real, dimension(:), allocatable						:: wavelet, wavelet_buff !Source wavelet 
  integer												:: wantwavefield, readwavefield, verbose
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right
  real, dimension(:,:,:), allocatable					:: sec_scale, field_tmp
  real, dimension(:,:,:,:), allocatable					:: source_wavefield, scat1_wavefield, scat2_wavefield
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold, ptemp
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: vel_pad, image1_pad, image2_pad  !Padded velocity model and images
  real													:: dt !Temporal sampling of the propagator
  real, dimension(:), allocatable						:: wavelet_fine, scaled_wavelet !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it, it2, itw !Cycles variables
  integer												:: iz, ix, iy !#######Cycles variables for fixing problem with ifort########
  integer												:: wc, srite, rc, sreed
  integer												:: desample, padded, vel_padded, sqslow
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr, total_time
  integer												:: count1, count2,ierr, icount, num_threads
  real*8, dimension(:,:), allocatable					:: interp_filter
  integer*8												:: rbit !necessary for reading big files
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  call from_par ("verbose",verbose,0)
		  

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)"####################Acoustic 2D/3D wave propagator#####################"
  if(verbose)write(0,*)"####################  Tomography forward operator  ####################"  
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("readwavefield",readwavefield,0)
  call from_par ("vel_padded",vel_padded,0)
  call from_par ("desample",desample,1)
  call from_par ("dt",dt,0.)
  call from_par ("padded",padded,0)
  call from_par ("sqslow",sqslow,0)
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the velocity
  if(vel_padded .eq. 0) then
  	allocate(vel(nz,nx,ny))
  	vel(:,:,:) = 0.
  	rc=sreed("in" , vel , nz*nx*ny*4)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	allocate(vel_pad(nz,nx,ny))
  	vel_pad(:,:,:) = 0.
  	rc=sreed("in" , vel_pad, nz*nx*ny*4)
  endif
  if(vel_padded .eq. 1 .and. padded .eq. 0) then
  	write(0,*)"ERROR WHEN VELOCITY PADDED, PERTURBATION MUST BE PADDED TOO"
  	stop
  endif
  !Adding absorbing boundaries to the velocity model and perturbation
  if (padded .eq. 0) then
  	!Not padded
  	allocate(image1(nz,nx,ny));image1(:,:,:) = 0.
  	allocate(image2(nz,nx,ny));image2(:,:,:) = 0.
  	rc=sreed("image1" , image1 , nz*nx*ny*4)
  	rc=sreed("image2" , image2 , nz*nx*ny*4)
  	call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	call Pad_perturbation(image1,image1_pad,nv_pad)
  	call Pad_perturbation(image2,image2_pad,nv_pad)
  else
  	!Padded
  	if(vel_padded .eq. 0) call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	allocate(image1_pad(nz,nx,ny));image1_pad(:,:,:) = 0.
  	rc=sreed("image1" , image1_pad , nz*nx*ny*4)
  	allocate(image2_pad(nz,nx,ny));image2_pad(:,:,:) = 0.
  	rc=sreed("image2" , image2_pad , nz*nx*ny*4)
  endif  
  

  !compute model perturbation from velocity perturbation
  if(sqslow .eq. 0) image1_pad(:,:,:) = -2.0*image1_pad(:,:,:)/(vel_pad(:,:,:)*vel_pad(:,:,:)*vel_pad(:,:,:))
  if(sqslow .eq. 0) image2_pad(:,:,:) = -2.0*image2_pad(:,:,:)/(vel_pad(:,:,:)*vel_pad(:,:,:)*vel_pad(:,:,:))
  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minval(vel_pad)/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxval(vel_pad)/min(dx,dy,dz)
  if(verbose)write(0,*) ""


  !Computing the second order derivative on the source wavelet
  call sec_deriv_field_init(nw, dw)

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_step_init(nz, nx, ny, nt, nv_pad, vel_pad, dt, dz, dx, dy,twoD)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,vel_pad,twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  old = 0.; cur = 0.
  pold => old
  pcur => cur
  pnew => old
  allocate(sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = -dt*dt*vel_pad*vel_pad
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !allocating memory for source wavefield
  allocate(source_wavefield(nz,nx,ny,nw),scat1_wavefield(nz,nx,ny,nw),scat2_wavefield(nz,nx,ny,nw))
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))
  allocate(field_tmp(nz,nx,ny))


  
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  

  if (wantwavefield) then
	if(verbose)write(0,*)"#Program will be writing wavefield as requested#"
  endif
  
  total_time = 0.
  rbit = nz*nx*ny
  rbit = rbit*nw*4
  do is = 1,ns
  	old = 0.; cur = 0.
  	source_wavefield = 0.; scat1_wavefield = 0.; scat2_wavefield = 0.
  	sec_source_left=0.; sec_source_right=0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	call system_clock(count_rate=cr)
	call system_clock(count1)
	
!##################################################################	

	if(.not. readwavefield) then
		!Scaling the source
		scaled_wavelet = wavelet_fine
		call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
		call Sou_scale_op(scaled_wavelet,vel_pad) !Adjoint = Forward (input-output same variable)
		if(verbose)write(0,*)"Propagating source wavefield..."
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
		endif
		do itw = 1, nw-1 !Forward
			it = (itw-1)*desample
			!Propagate source field	
			do it2 = 1, desample
				call Prop_step_op(.false., pold, pcur, pnew)
				!Adding source (perturbed and source wavefield)
				if (twoD) then
					call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				else
					call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				endif	
				!Apply adjoint(fine to coarse) of linear interpolation
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,itw) = source_wavefield(iz,ix,iy,itw) + pcur(iz,ix,iy)*interp_filter(it2,1)
							source_wavefield(iz,ix,iy,itw+1) = source_wavefield(iz,ix,iy,itw+1) + pcur(iz,ix,iy)*interp_filter(it2,2)	
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
				!Swapping pointers
				pold => pcur 
				pcur => pnew
				pnew => pold
			enddo
			!saving source wavefield
			if (wantwavefield) then
				wc=srite("wavefields",source_wavefield(:,:,:,itw),nz*nx*ny*4)
			endif
		
			if (mod(itw,(nw/10))==0) then
        		if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        		close(0)
      		endif
      	
		enddo
		!Apply adjoint(fine to coarse) of linear interpolation
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,nw) = source_wavefield(iz,ix,iy,nw) + pcur(iz,ix,iy)*interp_filter(1,1)
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO
		if (wantwavefield) then
			wc=srite("wavefields",source_wavefield(:,:,:,nw),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
		if(verbose)write(0,'(a)',advance="yes")""
		if(verbose)write(0,*)"End Propagation source wavefield"
	else	
		if(verbose)write(0,*)"Reading source wavefield"
		rc=sreed("wavefields",source_wavefield,rbit)
	endif
	
!##################################################################	
	
	!Scattered source wavefield on image1
	if(verbose)write(0,*)"Propagating scattered source wavefield on image1..."
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
  	sec_source_left=0.; sec_source_right=0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
  	
  	!Apply the adjoint of imaging condition (i.e., Born operator)
	!left sample source wavefield
	call sec_deriv_field_op(source_wavefield, field_tmp, 1)
	field_tmp(:,:,:) = field_tmp(:,:,:) * image1_pad(:,:,:)
	!Scale secondary source
	call weight_sec_op(.false., .false., field_tmp(:,:,:)*sec_scale(:,:,:), psec_source_left(:,:,:))
  	
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Apply the adjoint of imaging condition (i.e., Born operator)
		!right sample source wavefield
		call sec_deriv_field_op(source_wavefield, field_tmp, itw+1)
		field_tmp(:,:,:) = field_tmp(:,:,:) * image1_pad(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., field_tmp(:,:,:)*sec_scale(:,:,:), psec_source_right(:,:,:))
		
		!Propagate scattered wavefield
		do it2 = 1, desample
			call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)

			!Apply adjoint(fine to coarse) of linear interpolation
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						scat1_wavefield(iz,ix,iy,itw) = scat1_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(it2,1)
						scat1_wavefield(iz,ix,iy,itw+1) = scat1_wavefield(iz,ix,iy,itw+1) + pcurpert(iz,ix,iy)*interp_filter(it2,2)	
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
			
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("scatfield1",scat1_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_right = 0.
		
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
	enddo
		itw = nw
		!Apply adjoint(fine to coarse) of linear interpolation
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						scat1_wavefield(iz,ix,iy,itw) = scat1_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(1,1)
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
		!saving last scattered wavefield step
		if (wantwavefield) then
			wc=srite("scatfield1",scat1_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
		close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	if(verbose)write(0,*)"End Propagation scattered source wavefield on image1"
	
	
!##################################################################	

	!Scattered source wavefield on image2
	if(verbose)write(0,*)"Propagating scattered source wavefield on image2..."
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
  	sec_source_left=0.; sec_source_right=0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
  	
  	!Apply the adjoint of imaging condition (i.e., Born operator)
	!left sample source wavefield
	call sec_deriv_field_op(source_wavefield, field_tmp, 1)
	field_tmp(:,:,:) = field_tmp(:,:,:) * image2_pad(:,:,:)
	!Scale secondary source
	call weight_sec_op(.false., .false., field_tmp(:,:,:)*sec_scale(:,:,:), psec_source_left(:,:,:))
  	
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Apply the adjoint of imaging condition (i.e., Born operator)
		!right sample source wavefield
		call sec_deriv_field_op(source_wavefield, field_tmp, itw+1)
		field_tmp(:,:,:) = field_tmp(:,:,:) * image2_pad(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., field_tmp(:,:,:)*sec_scale(:,:,:), psec_source_right(:,:,:))
		
		!Propagate scattered wavefield
		do it2 = 1, desample
			call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)

			!Apply adjoint(fine to coarse) of linear interpolation
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						scat2_wavefield(iz,ix,iy,itw) = scat2_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(it2,1)
						scat2_wavefield(iz,ix,iy,itw+1) = scat2_wavefield(iz,ix,iy,itw+1) + pcurpert(iz,ix,iy)*interp_filter(it2,2)	
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
			
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("scatfield2",scat2_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_right = 0.
		
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
	enddo
		itw = nw
			!Apply adjoint(fine to coarse) of linear interpolation
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						scat2_wavefield(iz,ix,iy,itw) = scat2_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(1,1)	
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO

		!saving last scattered wavefield step
		if (wantwavefield) then
			wc=srite("scatfield2",scat2_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
		close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	if(verbose)write(0,*)"End Propagation scattered source wavefield on image2"

	
!##################################################################	
	!Double-scattering propagation with scattering on image1 and image2
	!Setting initial pointers
	if(verbose)write(0,*)"Propagating double-scattered wavefield..."
	old = 0.; cur = 0.
	poldpert => old
  	pcurpert => cur
  	pnewpert => old
  	if(wantwavefield) source_wavefield = 0.; !Only in case scattered wavefield is requested
  	sec_source_left=0.; sec_source_right=0.
  	psec_source_left=>sec_source_left
  	psec_source_right=>sec_source_right
  	
  	!left sample secondary source scat1 on image2
	call sec_deriv_field_op(scat1_wavefield, field_tmp, 1)
	field_tmp(:,:,:) = field_tmp(:,:,:) * image2_pad(:,:,:) * sec_scale(:,:,:)
	!Scale secondary source
	call weight_sec_op(.false., .false., field_tmp(:,:,:), psec_source_left(:,:,:))
	!left sample secondary source scat2 on image1
	call sec_deriv_field_op(scat2_wavefield, field_tmp, 1)
	field_tmp(:,:,:) = field_tmp(:,:,:) * image1_pad(:,:,:) * sec_scale(:,:,:)
	!Scale secondary source
	call weight_sec_op(.false., .true., field_tmp(:,:,:), psec_source_left(:,:,:))

	!Scattered wavefield
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Apply the adjoint of imaging condition (i.e., Born operator)
		!right sample secondary source scat1 on image2
		call sec_deriv_field_op(scat1_wavefield, field_tmp, itw+1)
		field_tmp(:,:,:) = field_tmp(:,:,:) * image2_pad(:,:,:) * sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., field_tmp(:,:,:), psec_source_right(:,:,:))
		!right sample secondary source scat1 on image2
		call sec_deriv_field_op(scat2_wavefield, field_tmp, itw+1)
		field_tmp(:,:,:) = field_tmp(:,:,:) * image1_pad(:,:,:) * sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .true., field_tmp(:,:,:), psec_source_right(:,:,:))
		
		!Propagate scattered wavefield
		do it2 = 1, desample
			call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)
			if (twoD) then
				!Taking perturbed wavefield
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			else
				!Taking perturbed wavefield
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			endif
			if (wantwavefield) then
				!Apply adjoint(fine to coarse) of linear interpolation
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy=1,ny
					do ix=1,nx
						do iz=1,nz
							source_wavefield(iz,ix,iy,itw) = source_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(it2,1)
							source_wavefield(iz,ix,iy,itw+1) = source_wavefield(iz,ix,iy,itw+1) + pcurpert(iz,ix,iy)*interp_filter(it2,2)	
						enddo
					enddo
				enddo
				!$OMP END PARALLEL DO		
			endif
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("wavefield",source_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		
		!Swap secondary tmp left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		psec_source_right = 0.
      	
	enddo	
		if (wantwavefield) then
			itw=nw
			!Apply adjoint(fine to coarse) of linear interpolation
			!$OMP PARALLEL DO private(iz,ix,iy)
			do iy=1,ny
				do ix=1,nx
					do iz=1,nz
						source_wavefield(iz,ix,iy,itw) = source_wavefield(iz,ix,iy,itw) + pcurpert(iz,ix,iy)*interp_filter(1,1)	
					enddo
				enddo
			enddo
			!$OMP END PARALLEL DO
			!saving last scattered wavefield step
			wc=srite("wavefield",source_wavefield(:,:,:,itw),nz*nx*ny*4)
		endif
		
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	
	if(verbose)write(0,'(a)',advance="yes")""
	if(verbose)write(0,*)"End Propagation double-scattered wavefield"
	
!##################################################################	
	
	
	call system_clock(count2)
	if(verbose)write(0,'(a)',advance="yes")""
	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	total_time = total_time + (count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing receiver wavefield for source n°: ", is
	!Resample receivers wavefield
	if (dw .le. dt) then
		do ig = 1,ng
			wc=srite("out",recfield_fine(:,ig),nw*4)
		enddo
	else
		do ig = 1,ng
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		enddo
	endif
   	if(verbose)write(0,*) ""

  enddo
  
  if(verbose)write(0,*) "Total propagation time: ", total_time/60.0 ," minutes"
  
  !Closing used routines
  call Prop_step_close()
  call weight_sec_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  
  
  if (wantwavefield) then
  
  	if(twoD) then
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
  			call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",nt/desample+1,"wavefields")
  			call to_history("o3",ow,"wavefields")
 		 	call to_history("d3",dt*desample,"wavefields")
  			call to_history("label3","Time [s]","wavefields")
  			call to_history("n4",ns,"wavefields")
  			call to_history("o4",1.,"wavefields")
  			call to_history("d4",1.,"wavefields")
  			call to_history("label4","Source number","wavefields") 
  		endif 
  		
  		
  			call to_history("n1",nz,"scatfield1")
  			call to_history("o1",oz,"scatfield1")
  			call to_history("d1",dz,"scatfield1")
  			call to_history("label1","z [m]","scatfield1")
			call to_history("n2",nx,"scatfield1")
		  	call to_history("o2",ox,"scatfield1")
  			call to_history("d2",dx,"scatfield1")
  			call to_history("label2","x [m]","scatfield1")
			call to_history("n3",nt/desample+1,"scatfield1")
  			call to_history("o3",ow,"scatfield1")
 		 	call to_history("d3",dt*desample,"scatfield1")
  			call to_history("label3","Time [s]","scatfield1")
  			call to_history("n4",ns,"scatfield1")
  			call to_history("o4",1.,"scatfield1")
  			call to_history("d4",1.,"scatfield1")
  			call to_history("label4","Source number","scatfield1") 
  			
  			call to_history("n1",nz,"scatfield2")
  			call to_history("o1",oz,"scatfield2")
  			call to_history("d1",dz,"scatfield2")
  			call to_history("label1","z [m]","scatfield2")
			call to_history("n2",nx,"scatfield2")
		  	call to_history("o2",ox,"scatfield2")
  			call to_history("d2",dx,"scatfield2")
  			call to_history("label2","x [m]","scatfield2")
			call to_history("n3",nt/desample+1,"scatfield2")
  			call to_history("o3",ow,"scatfield2")
 		 	call to_history("d3",dt*desample,"scatfield2")
  			call to_history("label3","Time [s]","scatfield2")
  			call to_history("n4",ns,"scatfield2")
  			call to_history("o4",1.,"scatfield2")
  			call to_history("d4",1.,"scatfield2")
  			call to_history("label4","Source number","scatfield2") 
  		
  		
  			call to_history("n1",nz,"scatfield12")
  			call to_history("o1",oz,"scatfield12")
  			call to_history("d1",dz,"scatfield12")
  			call to_history("label1","z [m]","scatfield12")
			call to_history("n2",nx,"scatfield12")
		  	call to_history("o2",ox,"scatfield12")
  			call to_history("d2",dx,"scatfield12")
  			call to_history("label2","x [m]","scatfield12")
			call to_history("n3",nt/desample+1,"scatfield12")
  			call to_history("o3",ow,"scatfield12")
 		 	call to_history("d3",dt*desample,"scatfield12")
  			call to_history("label3","Time [s]","scatfield12")
  			call to_history("n4",ns,"scatfield12")
  			call to_history("o4",1.,"scatfield12")
  			call to_history("d4",1.,"scatfield12")
  			call to_history("label4","Source number","scatfield12") 		
  	else
  		if(.not. readwavefield) then
			call to_history("n1",nz,"wavefields")
  			call to_history("o1",oz,"wavefields")
  			call to_history("d1",dz,"wavefields")
  			call to_history("label1","z [m]","wavefields")
			call to_history("n2",nx,"wavefields")
		  	call to_history("o2",ox,"wavefields")
	  		call to_history("d2",dx,"wavefields")
  			call to_history("label2","x [m]","wavefields")
			call to_history("n3",ny,"wavefields")
		  	call to_history("o3",oy,"wavefields")
		  	call to_history("d3",dy,"wavefields")
  			call to_history("label3","y [m]","wavefields")
			call to_history("n4",nt/desample+1,"wavefields")
  			call to_history("o4",ow,"wavefields")
 		 	call to_history("d4",dt*desample,"wavefields")
  			call to_history("label4","Time [s]","wavefields")
  			call to_history("n5",ns,"wavefields")
  			call to_history("o5",1.,"wavefields")
  			call to_history("d5",1.,"wavefields")
  			call to_history("label5","Source number","wavefields")
  		endif
  			
  			call to_history("n1",nz,"scatfield1")
  			call to_history("o1",oz,"scatfield1")
  			call to_history("d1",dz,"scatfield1")
  			call to_history("label1","z [m]","scatfield1")
			call to_history("n2",nx,"scatfield1")
		  	call to_history("o2",ox,"scatfield1")
  			call to_history("d2",dx,"scatfield1")
  			call to_history("label2","x [m]","scatfield1")
			call to_history("n3",ny,"scatfield1")
		  	call to_history("o3",oy,"scatfield1")
		  	call to_history("d3",dy,"scatfield1")
  			call to_history("label3","y [m]","scatfield1")
			call to_history("n4",nt/desample+1,"scatfield1")
  			call to_history("o4",ow,"scatfield1")
 	 		call to_history("d4",dt*desample,"scatfield1")
  			call to_history("label4","Time [s]","scatfield1")
  			call to_history("n5",ns,"scatfield1")
  			call to_history("o5",1.,"scatfield1")
  			call to_history("d5",1.,"scatfield1")
  			call to_history("label5","Source number","scatfield1")
  			
  			call to_history("n1",nz,"scatfield2")
  			call to_history("o1",oz,"scatfield2")
  			call to_history("d1",dz,"scatfield2")
  			call to_history("label1","z [m]","scatfield2")
			call to_history("n2",nx,"scatfield2")
		  	call to_history("o2",ox,"scatfield2")
  			call to_history("d2",dx,"scatfield2")
  			call to_history("label2","x [m]","scatfield2")
			call to_history("n3",ny,"scatfield2")
		  	call to_history("o3",oy,"scatfield2")
		  	call to_history("d3",dy,"scatfield2")
  			call to_history("label3","y [m]","scatfield2")
			call to_history("n4",nt/desample+1,"scatfield2")
  			call to_history("o4",ow,"scatfield2")
 	 		call to_history("d4",dt*desample,"scatfield2")
  			call to_history("label4","Time [s]","scatfield2")
  			call to_history("n5",ns,"scatfield2")
  			call to_history("o5",1.,"scatfield2")
  			call to_history("d5",1.,"scatfield2")
  			call to_history("label5","Source number","scatfield2")
  		
  		
  			call to_history("n1",nz,"scatfield12")
  			call to_history("o1",oz,"scatfield12")
  			call to_history("d1",dz,"scatfield12")
  			call to_history("label1","z [m]","scatfield12")
			call to_history("n2",nx,"scatfield12")
		  	call to_history("o2",ox,"scatfield12")
  			call to_history("d2",dx,"scatfield12")
  			call to_history("label2","x [m]","scatfield12")
			call to_history("n3",ny,"scatfield12")
		  	call to_history("o3",oy,"scatfield12")
		  	call to_history("d3",dy,"scatfield12")
  			call to_history("label3","y [m]","scatfield12")
			call to_history("n4",nt/desample+1,"scatfield12")
  			call to_history("o4",ow,"scatfield12")
 	 		call to_history("d4",dt*desample,"scatfield12")
  			call to_history("label4","Time [s]","scatfield12")
  			call to_history("n5",ns,"scatfield12")
  			call to_history("o5",1.,"scatfield12")
  			call to_history("d5",1.,"scatfield12")
  			call to_history("label5","Source number","scatfield12")	
  	endif
  endif
  if(verbose)write(0,*) "End propagation"

  !Deallocating memory
  deallocate(field_tmp)
  deallocate(interp_filter,scat2_wavefield,scat1_wavefield,source_wavefield)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale)
  deallocate(cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(image2_pad,image1_pad,vel_pad)
  deallocate(wavelet)
  !deallocate(wavelet_buff,wavelet)
  call sep_close()
  call exit (0)
end program Hessian_tomo

