!Module for computing the Laplacian with 10th-order approximation (zero-outside -> Forward=Adjoint)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput laplacian of wavefield
!D(z,x,y) = Laplacian(Model(z,x,y))
!Model(z,x,y) = Laplacian*(D(z,x,y))
module Laplacian

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  real, private									:: dz, dx, dy
  real, private 								:: C0
  real, private 								:: Cz0, Cz1, Cz2, Cz3, Cz4, Cz5
  real, private 								:: Cx0, Cx1, Cx2, Cx3, Cx4, Cx5
  real, private 								:: Cy0, Cy1, Cy2, Cy3, Cy4, Cy5
  real, private, dimension(:,:,:),allocatable	:: tmp
  logical, private								:: twoD

  contains


	!Initialization routine
	subroutine Laplacian_init(nz_in, nx_in, ny_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real, intent(in)						:: dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		!Internal variables
		integer 								:: buffer

   		parameter(buffer=5)

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dz = dz_in
		dx = dx_in
		dy = dy_in
		twoD = twoD_in

	        if (.not. allocated(tmp)) then
 				if (.not. twoD) then
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer))	
				else
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1))	
				endif
				!Initializing 10th-order FD second-order derivative
				Cz0 = -2.927222222222182   /(dz*dz)
				Cz1 =  1.666666666666642   /(dz*dz)
				Cz2 = -0.238095238095235   /(dz*dz)
				Cz3 =  0.039682539682539   /(dz*dz)
				Cz4 = -0.004960317460317   /(dz*dz)
				Cz5 =  3.174603174603100e-4/(dz*dz)

				Cx0 = -2.927222222222182   /(dx*dx)
				Cx1 =  1.666666666666642   /(dx*dx)
				Cx2 = -0.238095238095235   /(dx*dx)
				Cx3 =  0.039682539682539   /(dx*dx)
				Cx4 = -0.004960317460317   /(dx*dx)
				Cx5 =  3.174603174603100e-4/(dx*dx)	
				C0 = Cx0 + Cz0
				if (.not. twoD) then
					Cy0 = -2.927222222222182   /(dy*dy)
					Cy1 =  1.666666666666642   /(dy*dy)	
					Cy2 = -0.238095238095235   /(dy*dy)
					Cy3 =  0.039682539682539   /(dy*dy)
					Cy4 = -0.004960317460317   /(dy*dy)
					Cy5 =  3.174603174603100e-4/(dy*dy)
					C0 = C0 + Cy0
				endif
			endif

		!Temporary variable for the Laplacian zeroing out
    		tmp = 0.0



		return
	end subroutine Laplacian_init

  	subroutine Laplacian_call(adj,add,mm,dd)
    	logical,intent(in)   :: adj,add
   		real,dimension(:) 	 :: mm,dd
    		call Laplacian_op(adj,add,mm,dd)
  	end subroutine Laplacian_call

	subroutine Laplacian_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm !Input 
		real, dimension(nz,nx,ny)			:: dd !Output 
		!Internal variables
		integer						:: it, iz, ix, iy


    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model

			 tmp(1:nz,1:nx,1:ny) = dd


			if(twoD) then
					!$OMP PARALLEL DO private(iz,ix)
			        do ix=1,nx
            			 do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) + C0 *tmp(iz,ix,1)         + &
                                      		       Cz1 * (tmp(iz+1,ix,1) + tmp(iz-1,ix,1)) + &
                                      		       Cz2 * (tmp(iz+2,ix,1) + tmp(iz-2,ix,1)) + &
                                      		       Cz3 * (tmp(iz+3,ix,1) + tmp(iz-3,ix,1)) + &
                                      		       Cz4 * (tmp(iz+4,ix,1) + tmp(iz-4,ix,1)) + &
                                      		       Cz5 * (tmp(iz+5,ix,1) + tmp(iz-5,ix,1)) + &
                                      		       Cx1 * (tmp(iz,ix+1,1) + tmp(iz,ix-1,1)) + &
                                          	       Cx2 * (tmp(iz,ix+2,1) + tmp(iz,ix-2,1)) + &
                                      		       Cx3 * (tmp(iz,ix+3,1) + tmp(iz,ix-3,1)) + &
                                       		       Cx4 * (tmp(iz,ix+4,1) + tmp(iz,ix-4,1)) + &
                                      		       Cx5 * (tmp(iz,ix+5,1) + tmp(iz,ix-5,1)) 
                		enddo
            		enddo
            		!$OMP END PARALLEL DO
			
			else

        	 !$OMP PARALLEL DO private(iz,ix,iy)
			 do iy = 1,ny
        		    do ix=1,nx
            			 do iz=1,nz
                			mm(iz,ix,iy) = mm(iz,ix,iy) + C0 *tmp(iz,ix,iy)         + &
                                      		       Cz1 * (tmp(iz+1,ix,iy) + tmp(iz-1,ix,iy)) + &
                                      		       Cz2 * (tmp(iz+2,ix,iy) + tmp(iz-2,ix,iy)) + &
                                      		       Cz3 * (tmp(iz+3,ix,iy) + tmp(iz-3,ix,iy)) + &
                                      		       Cz4 * (tmp(iz+4,ix,iy) + tmp(iz-4,ix,iy)) + &
                                      		       Cz5 * (tmp(iz+5,ix,iy) + tmp(iz-5,ix,iy)) + &
                                      		       Cx1 * (tmp(iz,ix+1,iy) + tmp(iz,ix-1,iy)) + &
                                          	       Cx2 * (tmp(iz,ix+2,iy) + tmp(iz,ix-2,iy)) + &
                                      		       Cx3 * (tmp(iz,ix+3,iy) + tmp(iz,ix-3,iy)) + &
                                       		       Cx4 * (tmp(iz,ix+4,iy) + tmp(iz,ix-4,iy)) + &
                                      		       Cx5 * (tmp(iz,ix+5,iy) + tmp(iz,ix-5,iy)) + &
                                      		       Cy1 * (tmp(iz,ix,iy+1) + tmp(iz,ix,iy-1)) + &
                                          	       Cy2 * (tmp(iz,ix,iy+2) + tmp(iz,ix,iy-2)) + &
                                      		       Cy3 * (tmp(iz,ix,iy+3) + tmp(iz,ix,iy-3)) + &
                                       		       Cy4 * (tmp(iz,ix,iy+4) + tmp(iz,ix,iy-4)) + &
                                      		       Cy5 * (tmp(iz,ix,iy+5) + tmp(iz,ix,iy-5))

                		enddo
            		enddo
			 enddo
        	 !$OMP END PARALLEL DO
        	 
			endif
		
		else	     !Forward
        		if (.not. add) dd = 0. !Zero-out the data

			 tmp(1:nz,1:nx,1:ny) = mm
			 
			 
			 
			if(twoD) then
					!$OMP PARALLEL DO private(iz,ix)
			        do ix=1,nx
            			 do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C0 *tmp(iz,ix,1)         + &
                                      		       Cz1 * (tmp(iz+1,ix,1) + tmp(iz-1,ix,1)) + &
                                      		       Cz2 * (tmp(iz+2,ix,1) + tmp(iz-2,ix,1)) + &
                                      		       Cz3 * (tmp(iz+3,ix,1) + tmp(iz-3,ix,1)) + &
                                      		       Cz4 * (tmp(iz+4,ix,1) + tmp(iz-4,ix,1)) + &
                                      		       Cz5 * (tmp(iz+5,ix,1) + tmp(iz-5,ix,1)) + &
                                      		       Cx1 * (tmp(iz,ix+1,1) + tmp(iz,ix-1,1)) + &
                                          	       Cx2 * (tmp(iz,ix+2,1) + tmp(iz,ix-2,1)) + &
                                      		       Cx3 * (tmp(iz,ix+3,1) + tmp(iz,ix-3,1)) + &
                                       		       Cx4 * (tmp(iz,ix+4,1) + tmp(iz,ix-4,1)) + &
                                      		       Cx5 * (tmp(iz,ix+5,1) + tmp(iz,ix-5,1)) 

                		enddo
            		enddo
            		!$OMP END PARALLEL DO
			
			else
  
        	 !$OMP PARALLEL DO private(iz,ix,iy)
			 do iy = 1,ny
        		    do ix=1,nx
            			 do iz=1,nz
                			dd(iz,ix,iy) = dd(iz,ix,iy) + C0 *tmp(iz,ix,iy)         + &
                                      		       Cz1 * (tmp(iz+1,ix,iy) + tmp(iz-1,ix,iy)) + &
                                      		       Cz2 * (tmp(iz+2,ix,iy) + tmp(iz-2,ix,iy)) + &
                                      		       Cz3 * (tmp(iz+3,ix,iy) + tmp(iz-3,ix,iy)) + &
                                      		       Cz4 * (tmp(iz+4,ix,iy) + tmp(iz-4,ix,iy)) + &
                                      		       Cz5 * (tmp(iz+5,ix,iy) + tmp(iz-5,ix,iy)) + &
                                      		       Cx1 * (tmp(iz,ix+1,iy) + tmp(iz,ix-1,iy)) + &
                                          	       Cx2 * (tmp(iz,ix+2,iy) + tmp(iz,ix-2,iy)) + &
                                      		       Cx3 * (tmp(iz,ix+3,iy) + tmp(iz,ix-3,iy)) + &
                                       		       Cx4 * (tmp(iz,ix+4,iy) + tmp(iz,ix-4,iy)) + &
                                      		       Cx5 * (tmp(iz,ix+5,iy) + tmp(iz,ix-5,iy)) + &
                                      		       Cy1 * (tmp(iz,ix,iy+1) + tmp(iz,ix,iy-1)) + &
                                          	       Cy2 * (tmp(iz,ix,iy+2) + tmp(iz,ix,iy-2)) + &
                                      		       Cy3 * (tmp(iz,ix,iy+3) + tmp(iz,ix,iy-3)) + &
                                       		       Cy4 * (tmp(iz,ix,iy+4) + tmp(iz,ix,iy-4)) + &
                                      		       Cy5 * (tmp(iz,ix,iy+5) + tmp(iz,ix,iy-5))

                		enddo
            		enddo
			 enddo
        	 !$OMP END PARALLEL DO

			endif
 
		endif

		
		return
	end subroutine Laplacian_op



end module Laplacian
