!!$=head1 NAME
!!$
!!$Propagator Forward - Acoustic propagator with constant density. (Version in which source wavefield interpolated with correct scale)
!!$
!!$=head1 SYNOPSIS
!!$
!!$Progator_fw <velocity.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H threads= >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation solver for acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model and return the data.H
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<velocity.H> - Velocity model in SEP3D format (theoretically [m/s]). (Constructed using Velpar and/or Vel)
!!$
!!$=item from par
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<verbose> - [0] If 1 print propagation information
!!$
!!$B<integer> I<vel_padded> - [0] Velocity model was padded (100 elements), useful for FWI
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Propagator
  
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_step


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: vel   !velocity model 
  real, dimension(:), allocatable						:: wavelet, scaled_wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads, verbose, vel_padded
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: vel_pad !Padded velocity model
  real													:: dt !Temporal sampling of the propagator
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it, iz,ix !Cycles variables
  integer												:: wc, srite, rc, sreed
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr, total_time
  integer												:: count1, count2,ierr
  !Variables to interpolate source wavefield as propagates
  real, dimension(:,:,:), allocatable, target			:: left_sou, right_sou
  real, dimension(:,:,:), pointer						:: pleft_sou, pright_sou, ptemp
  real, dimension(:,:), allocatable						:: interp_filter
  integer												:: it2, itw
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  call from_par ("verbose",verbose,0)
		  
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)"####################Acoustic 2D/3D wave propagator#####################"
  if(verbose)write(0,*)"####################            Forward           #####################"  
  if(verbose)write(0,*)"#######################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("vel_padded",vel_padded,0)
  call from_par ("dt",dt,0.)
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the velocity
  if(vel_padded .eq. 0) then
  	allocate(vel(nz,nx,ny))
  	vel(:,:,:) = 0.
  	rc=sreed("in" , vel , nz*nx*ny*4)
  	!Adding absorbing boundaries to the velocity model
  	call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	allocate(vel_pad(nz,nx,ny))
  	vel_pad(:,:,:) = 0.
  	rc=sreed("in" , vel_pad, nz*nx*ny*4)
  endif

  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minval(vel_pad)/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxval(vel_pad)/min(dx,dy,dz)
  if(verbose)write(0,*)"	Min vel [m/s]:",minval(vel_pad)
  if(verbose)write(0,*)"	Max vel [m/s]:",maxval(vel_pad)
  if(verbose)write(0,*) ""
  
  

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)

  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_step_init(nz, nx, ny, nt, nv_pad, vel_pad, dt, dz, dx, dy,twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny),left_sou(nz,nx,ny),right_sou(nz,nx,ny))
  left_sou = 0.
  right_sou = 0.
  pleft_sou => left_sou
  pright_sou => right_sou
  pold => old
  pcur => cur
  pnew => old
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))

  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  
  if (wantwavefield) then 
  	if(verbose)write(0,*) "#Program will be writing wavefield as requested#"
  endif

	
  total_time = 0.
  do is = 1,ns
  	old = 0.; cur = 0.
  	left_sou = 0.; right_sou = 0.
	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the source
	scaled_wavelet = wavelet_fine
	call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_op(scaled_wavelet,vel_pad) !Adjoint = Forward (input-output same variable)
	call system_clock(count_rate=cr)
	call system_clock(count1)
	if(verbose)write(0,*)"Propagating..."
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Propagate source field
		do it2 = 1, desample
			call Prop_step_op(.false., pold, pcur, pnew)
			!Adding source and Taking wavefield
			if (twoD) then
				call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
				call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.false.,.false.,pcur,recfield_fine(it+it2,ig))
				enddo
			else
				call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
				call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.false.,.false.,pcur,recfield_fine(it+it2,ig))
				enddo
			endif
			!Apply adjoint(fine to coarse) of linear interpolation
			if (wantwavefield) then
				pleft_sou(:,:,:) = pleft_sou(:,:,:) + pcur*interp_filter(it2,1)
				pright_sou(:,:,:) = pright_sou(:,:,:) + pcur*interp_filter(it2,2)
			endif
			!Swapping pointers
			pold => pcur 
			pcur => pnew
			pnew => pold
		enddo
		!saving source wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pleft_sou(:,:,:),nz*nx*ny*4)
		endif
		
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	!Swap source wavefield left and right pointers
		ptemp => pleft_sou
		pleft_sou => pright_sou
		pright_sou => ptemp
		pright_sou = 0.
	enddo
		!Last step of the source wavefield just saving source wavefield
		if (wantwavefield) then
			pleft_sou(:,:,:) = pleft_sou(:,:,:) + pcur*interp_filter(1,1)
			wc=srite("wavefield",pleft_sou(:,:,:),nz*nx*ny*4)
		endif
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
    if(verbose)write(0,'(a)',advance="yes")""
	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	total_time = total_time + (count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing receiver wavefield for source n°: ", is
	!Resample receivers' wavefield
	do ig = 1,ng
		if (dw .le. dt) then
			wc=srite("out",recfield_fine(:,ig),nw*4)
		else
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		endif
	enddo
   	if(verbose)write(0,*) ""
  enddo
  
  if(verbose)write(0,*) "Total propagation time: ", total_time ," seconds"
  if(verbose)write(0,*) "Total propagation time: ", total_time/60.0 ," minutes"

  !Closing used routines
  call Prop_step_close()
  call Sou_scale_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",nt/desample+1,"wavefield")
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  		call to_history("n4",ns,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Source number","wavefield")  	
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		call to_history("n4",nt/desample+1,"wavefield")
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"

  !Deallocating memory
  deallocate(interp_filter)
  deallocate(right_sou,left_sou,cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(vel_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)
end program Propagator


