!Module for computing the weight operator (see Ali's note) w (weight in the boundary region)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput wavefield
!D(z,x,y) = weight(Model(z,x,y))
!Model(z,x,y) = weight*(D(z,x,y))
!NOTE ADJOINT = FORWARD
module weight

  implicit none
  !Module variables
  integer, private									:: nz, nx, ny, npad
  real, private										:: dt
  real, dimension(:,:,:), allocatable, private		:: wz, wx, wy
  logical, private									:: twoD

  contains


	!Initialization routine
	subroutine weight_init(nz_in, nx_in, ny_in, dt_in, dz_in, dx_in, dy_in, nv_pad_in, vel_in,twoD_in)
		
		!Input variables	
		integer, intent(in)					:: nz_in, nx_in, ny_in, nv_pad_in
		real								:: dt_in, dz_in, dx_in, dy_in
		real, dimension(:,:,:), pointer		:: vel_in
		logical, intent(in)					:: twoD_in
		!Internal variables	
		integer								:: i


		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		dt = dt_in
		npad = nv_pad_in
		twoD = twoD_in
	   	if (.not. allocated(wz)) allocate(wz(npad*2,nx,ny))
	   	if (.not. allocated(wx)) allocate(wx(nz,npad*2,ny))
    	wz  = 0.; wx = 0.

		!Z-axis weight
		!$OMP PARALLEL DO private(i)
		do i=1,npad
		        wz(i,:,:)			=  dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(i,:,:) * dt/dz_in
        		wz(2*npad-i+1,:,:)	=  dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(nz-i+1,:,:) * dt/dz_in
		enddo
		!$OMP END PARALLEL DO

		!X-axis weight
		!$OMP PARALLEL DO private(i)
		do i=1,npad
		        wx(:,i,:)			= dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(:,i,:) * dt/dx_in
        		wx(:,2*npad-i+1,:)	= dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(:,nx-i+1,:) * dt/dx_in
		enddo
		!$OMP END PARALLEL DO

		if(.not. twoD) then
			if (.not. allocated(wy)) allocate(wy(nz,nx,npad*2))
			wy = 0.
			!Y-axis weight	
			!$OMP PARALLEL DO private(i)
			do i=1,npad
			        wy(:,:,i)			= dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(:,:,i) * dt/dy_in
       		 		wy(:,:,2*npad-i+1)	= dfloat(npad-i+1)/dfloat(npad) * 0.15 * vel_in(:,:,ny-i+1) * dt/dy_in
			enddo
			!$OMP END PARALLEL DO
		endif

		return
	end subroutine weight_init


  	subroutine weight_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call weight_op(adj,add,mm,dd)
  	end subroutine weight_call

	
	!Operator routine
	subroutine weight_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)					:: adj, add
		real, dimension(nz,nx,ny)			:: mm 
		real, dimension(nz,nx,ny)			:: dd 
		!Internal variables
		integer								:: i


    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
				!Z-axis weighting
		        mm(1:npad,:,:) 			= mm(1:npad,:,:) 	   + dd(1:npad,:,:)*wz(1:npad,:,:)
        		mm(nz-npad+1:nz,:,:) 	= mm(nz-npad+1:nz,:,:) + dd(nz-npad+1:nz,:,:)*wz(npad+1:2*npad,:,:)  
				!X-axis weighting
		        mm(:,1:npad,:) 			= mm(:,1:npad,:) 	 	+ dd(:,1:npad,:)*wx(:,1:npad,:)
        		mm(:,nx-npad+1:nx,:) 	= mm(:,nx-npad+1:nx,:)  + dd(:,nx-npad+1:nx,:)*wx(:,npad+1:2*npad,:)  
				if(.not. twoD) then
					!Y-axis weighting
		        	mm(:,:,1:npad) 			= mm(:,:,1:npad) 	 	+ dd(:,:,1:npad)*wy(:,:,1:npad)
        			mm(:,:,ny-npad+1:ny) 	= mm(:,:,ny-npad+1:ny)  + dd(:,:,ny-npad+1:ny)*wy(:,:,npad+1:2*npad)  
				endif
			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				!Z-axis weighting
		        dd(1:npad,:,:) 			= dd(1:npad,:,:) 	   + mm(1:npad,:,:)*wz(1:npad,:,:)
        		dd(nz-npad+1:nz,:,:) 	= dd(nz-npad+1:nz,:,:) + mm(nz-npad+1:nz,:,:)*wz(npad+1:2*npad,:,:)  
				!X-axis weighting
		        dd(:,1:npad,:) 			= dd(:,1:npad,:) 	 	+ mm(:,1:npad,:)*wx(:,1:npad,:)
        		dd(:,nx-npad+1:nx,:) 	= dd(:,nx-npad+1:nx,:)  + mm(:,nx-npad+1:nx,:)*wx(:,npad+1:2*npad,:)  
				if(.not. twoD) then
					!Y-axis weighting
		        	dd(:,:,1:npad) 			= dd(:,:,1:npad) 	 	+ mm(:,:,1:npad)*wy(:,:,1:npad)
        			dd(:,:,ny-npad+1:ny) 	= dd(:,:,ny-npad+1:ny)  + mm(:,:,ny-npad+1:ny)*wy(:,:,npad+1:2*npad)  
				endif
			endif
		
		return
	end subroutine weight_op


	!Subroutine for deallocating variables
	subroutine weight_close()
		if (allocated(wy)) deallocate(wy)
		if (allocated(wx)) deallocate(wx)
		if (allocated(wz)) deallocate(wz)
	end subroutine weight_close


end module weight
